/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.api;

import org.objectweb.medor.clone.api.Cloneable;

/**
 * A TupleStructure is the basic data structure manipulated when expressing,
 * optimizing and evaluating MEDOR queries.
 * <p>A TupleStructure is composed of typed, named fields.
 * <p>Field names are unique within a given TupleStructure.
 * @author Sebastien Chassande-Barrioz
 * @see Field
 */
public interface TupleStructure extends Cloneable {

	/**
	 * Returns the array of Fields in the current TupleStructure.
	 * @return the array of Fields in the current TupleStructure.
	 */
    Field[] getFields();

	/**
	 * Returns a Field object present in this TupleStructure given its name.
	 * @param fieldname the name of the searched Field
	 * @return the Field in the current TupleStructure which has the input name.
	 * @throws MedorException if no Field corresponds to the input name.
	 */
    Field getField(String fieldname) throws MedorException;

	/**
	 * Returns the Field of a given rank in the current TupleStructure.
	 * @param fieldrank the rank of the searched Field in the current
	 * TupleStructure
	 * @return the Field in the current TupleStructure corresponding to the
	 * input rank.
	 * @throws MedorException if the input rank is too large.
	 */
    Field getField(int fieldrank) throws MedorException;

    /**
	 * Returns the rank of a given Field in the current TupleStructure.
	 * @param f the Field for which the rank is searched.
	 * @return the rank of the input Field in the current TupleStructure.
	 * @throws MedorException if there is no such Field in the current
	 * TupleStructure.
	 */
	int getFieldRank(Field f) throws MedorException;

	/**
	 * Returns the size of the current TupleStructure, ie the number of
	 * Fields.
	 * @return the number of Fields in the current TupleStructure.
	 */
    int getSize();

	/**
	 * Checks whether a given input Field is present in the current
	 * TupleStructure.
	 * @param f the Field for which to check whether it is present in the
	 * current TupleStructure.
	 * @return true if the Field is part of the current TupleStructure, false
	 * otherwise.
	 */
	boolean contains(Field f);

	/**
	 * Checks whether the current TupleStructure has a field of a given name.
	 * @param fieldName the Field name for which to check whether the current
	 * TupleStructure contains a Field.
	 * @return true if the current TupleStructure contains a Field of the input
	 * name, false otherwise.
	 */
	boolean contains(String fieldName);
}
