/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.medor.datasource.rdb.lib;

import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * This class encapsulates a resultset as a TupleCollection.
 *
 * @author S.Chassande-Barrioz
 */
public class JDBCTupleCollection implements TupleCollection {

	private ResultSet resultSet;
	private TupleStructure tupleStructure;
	private VariableOperand[] operandBuffer;
	private Tuple buffer;
	private RdbAdapter adapter;
	private PreparedStatement preparedStatement;
	private boolean debug;
	private Logger log;
	private int index;
	private boolean isLast = false;

	/**
	 * @param tupleStructure is the TupleStructure describing the struture
	 * of the TupleCollection.
	 * @param rs is the encapsulated ResultSet. The resultset is already
	 * positioned on the first element.
	 * @param ps is the PreparedStatement which must be closed in same time than
	 * this tuple collection and the ResultSet.
	 * @param adapter is the RdbAdapter to use
	 * @param logger
	 */
	public JDBCTupleCollection(TupleStructure tupleStructure,
							   ResultSet rs,
							   PreparedStatement ps,
							   RdbAdapter adapter,
							   Logger logger)
			throws MedorException, ExpressionException, SQLException {
		log = logger;
		debug = log.isLoggable(BasicLevel.DEBUG);
		if (debug) log.log(BasicLevel.DEBUG, "JDBCTupleCollection creation");
		this.tupleStructure = tupleStructure;
		resultSet = rs;
		int size = tupleStructure.getSize();
		// create and init the buffer
		operandBuffer = new VariableOperand[size];
		for (int cpt = 0; cpt < size; cpt++) {
			operandBuffer[cpt] = new BasicVariableOperand(
					tupleStructure.getField(cpt + 1).getType());
		}
		buffer = new MemoryTuple(operandBuffer);
		this.adapter = adapter;
		preparedStatement = ps;
		index = 0;
		isLast = false;
		next();
	}

	public TupleStructure getMetaData() throws MedorException {
		return tupleStructure;
	}

	/**
	 * Checks whether the current row is the last row.
	 * <p>This method relies on the resultSet.next() method, and not on the
	 * resultSet.isLast() method, since isLast is not supported by all JDBC
	 * drivers. Method next() of this class first checks whether
	 * resultSet.next() has already been called by isLast before calling it.</p>
	 * @return true if it was the last row, false otherwise
	 * @throws MedorException if there is a SQL exception
	 */
	public boolean isLast() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (debug) log.log(BasicLevel.DEBUG, "");
		return isLast;
	}

	public void close() throws MedorException {
		if (resultSet != null) {
			synchronized (this) {
				if (resultSet != null) {
					try {
						resultSet.close();
						preparedStatement.close();
					}
					catch (SQLException e) {
						throw new MedorException("Impossible to close the result set: ", e);
					}
					finally {
						resultSet = null;
						preparedStatement = null;
						tupleStructure = null;
						buffer = null;
						adapter = null;
						operandBuffer = null;
					}
				}
			}
		}
	}

	/**
	 * Moves the cursor down one row from its current position to the next
	 * row.
	 * <p>Note that nothing is done in the case resultSet.next() was already
	 * called when calling isLast().
	 * @return true if there was a next row, false if there is no next row.
	 * @throws MedorException if there is a SQL Exception.
	 */
	public boolean next() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (isLast) {
			if (debug) log.log(BasicLevel.DEBUG, "No more result");
			return false;
		}
		try {
			loadTuple();
			index ++;
			//compute the isLast value by calling the resultSet 
			isLast = !resultSet.next();
			if (debug) {
				log.log(BasicLevel.DEBUG, "Current tuple load from the resultset, index="
					+ index + ", isLast=" + isLast);
			}
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		} catch (ExpressionException sqlex) {
			throw new MedorException(sqlex);
		}
		return true;
	}

	/**
	 * Moves the cursor to the first Tuple of this TupleCollection
	 */
	public void first() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (debug) log.log(BasicLevel.DEBUG, "");
		if (index == 1) {
			return;
		}
		try {
			if (resultSet.first()) {
				index = 0;
				isLast = false;
				next();
				if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=true");
			} else
				throw new MedorException("can't move the cursor to the first row");
		} catch (SQLException ex) {
			throw new MedorException(ex);
		}
	}

	/**
	 * Retrieves the current row Number of this TupleCollection
	 */
	public int getRow() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (debug) log.log(BasicLevel.DEBUG, "");
		return index;
	}

	public Tuple getTuple() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (debug) log.log(BasicLevel.DEBUG, "");
		return buffer;
	}

	private void loadTuple()
			throws ExpressionException, MedorException, SQLException {
		PType ptype = null;
		int size = tupleStructure.getSize();
		StringBuffer sb;
		if (debug) {
			sb = new StringBuffer();
		}
		for (int cpt = 1; cpt <= size; cpt++) {
			ptype = tupleStructure.getField(cpt).getType();
			switch (ptype.getTypeCode()) {
			case QType.TYPECODE_INT:
				operandBuffer[cpt - 1].setValue(adapter.getInt(resultSet, cpt, -1));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_SHORT:
				operandBuffer[cpt - 1].setValue(adapter.getShort(resultSet, cpt, (short) -1));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_BYTE:
				operandBuffer[cpt - 1].setValue(adapter.getByte(resultSet, cpt, (byte) -1));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_LONG:
				operandBuffer[cpt - 1].setValue(adapter.getLong(resultSet, cpt, -1));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_DOUBLE:
				operandBuffer[cpt - 1].setValue(adapter.getDouble(resultSet, cpt, -1));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_BOOLEAN:
				operandBuffer[cpt - 1].setValue(adapter.getBoolean(resultSet, cpt, false));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_FLOAT:
				operandBuffer[cpt - 1].setValue(adapter.getFloat(resultSet, cpt, 0));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_STRING:
				operandBuffer[cpt - 1].setValue(adapter.getString(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJBOOLEAN:
				operandBuffer[cpt - 1].setValue(adapter.getOboolean(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJCHAR:
				operandBuffer[cpt - 1].setValue(adapter.getOchar(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJBYTE:
				operandBuffer[cpt - 1].setValue(adapter.getObyte(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJSHORT:
				operandBuffer[cpt - 1].setValue(adapter.getOshort(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJINT:
				operandBuffer[cpt - 1].setValue(adapter.getOint(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJLONG:
				operandBuffer[cpt - 1].setValue(adapter.getOlong(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJFLOAT:
				operandBuffer[cpt - 1].setValue(adapter.getOfloat(resultSet, cpt, null));
				break;
			case QType.TYPECODE_OBJDOUBLE:
				operandBuffer[cpt - 1].setValue(adapter.getOdouble(resultSet, cpt, null));
				break;
			case QType.TYPECODE_DATE:
				operandBuffer[cpt - 1].setValue(adapter.getDate(resultSet, cpt, null));
				break;
			case QType.TYPECODE_CHAR:
				operandBuffer[cpt - 1].setValue(adapter.getChar(resultSet, cpt, (char) 0));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_BYTEARRAY:
				operandBuffer[cpt - 1].setValue(adapter.getByteArray(resultSet, cpt, null));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_CHARARRAY:
				operandBuffer[cpt - 1].setValue(adapter.getCharArray(resultSet, cpt, null));
				operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
				break;
			case QType.TYPECODE_BIGDECIMAL:
				operandBuffer[cpt - 1].setValue(adapter.getBigDecimal(resultSet, cpt, null));
				break;
			case QType.TYPECODE_BIGINTEGER:
				operandBuffer[cpt - 1].setValue(adapter.getBigInteger(resultSet, cpt, null));
				break;
			case PType.TYPECODE_SERIALIZED:
				try {
					operandBuffer[cpt - 1].setValue(adapter.getSerialized(resultSet, cpt, null));
				}
				catch (Exception e) {
					throw new MedorException("Impossible to fetch a serialized field a the index " + cpt, e);
				}
				break;
			}
		}
		if (debug) {
			sb = new StringBuffer("[");
			String sep = "";
			for (int cpt = 0; cpt < size; cpt++) {
				sb.append(sep);
				sb.append(operandBuffer[cpt]);
				sep = ",";
			}
			sb.append(']');
			log.log(BasicLevel.DEBUG, "Tuple loaded: " + sb.toString());
		}

	}

	public Tuple getTuple(int row) throws MedorException {
		if (row(row)) {
			return getTuple();
		} else {
			throw new MedorException("Can't move to the row: " + row);
		}
	}

	public boolean isEmpty() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (debug) log.log(BasicLevel.DEBUG, "return " + (buffer == null));
		return (buffer == null);
	}

	public boolean row(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		if (index == i) {
			return true;
		}
		try {
			if (debug) log.log(BasicLevel.DEBUG, "resultSet.absolute(" + i + ")");
			if (resultSet.absolute(i)) {
				index = i-1;
				isLast = false;
				next();
				return true;
			} else
				return false;
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		}
	}
}


