/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.datasource.rdb.lib;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.BasicRdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.OracleAdapter;
import org.objectweb.jorm.mapper.rdb.lib.RdbConnectionWrapper;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.expression.api.CalculatedParameterOperand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.InCollection;
import org.objectweb.medor.filter.lib.IsEmpty;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.query.rdb.lib.AggregateRdbQueryNode;
import org.objectweb.medor.query.rdb.lib.MedorExpression2WhereClause;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.EmptyTupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.eval.api.EvaluationMetaData;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.Serializable;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

/**
 * This class implements the Wrapper interface for the jdbc resources. This
 * implementation uses Statements to perform the request.
 *
 * @author Mourad Alia
 */
public class JDBCWrapper implements Wrapper {

	/**
	 * this field references a Logger which is never null.
	 */
	static Logger log = null;
	static Logger TClog = null;
	boolean debug;

	public JDBCWrapper() {
		if (log == null) {
			log = Log.loggerFactory.getLogger(
					Log.MEDOR_PREFIX + ".eval.rdb.JDBCWrapper");
			TClog = Log.loggerFactory.getLogger(
					Log.MEDOR_PREFIX + ".eval.rdb.JDBCTC");
		}
	}

	public TupleCollection fetchData(QueryLeaf lf,
									 ParameterOperand[] parameters,
									 Object connection,
									 EvaluationMetaData evalMD)
			throws EvaluationException {
		debug = log.isLoggable(BasicLevel.DEBUG);
		if (debug) {
			log.log(BasicLevel.DEBUG, "Parameters: " + posToString(parameters));
		}
		RdbQueryLeaf leaf = (RdbQueryLeaf) lf;
		//Calculate the parameter list of the leaf with the parameter list of the query
		RdbAdapter rdbAdapt = getRdbAdapter(leaf);
		Connection conn = getConnection(connection);
		ResultSet resultSet = null;
		PreparedStatement ps = null;
		String sqlRequest;
        ParameterOperand[] usedParameters = getUsedPOs(leaf, parameters, rdbAdapt, evalMD);
		synchronized (leaf) {
			try {
				sqlRequest = leaf.getSqlRequest(parameters,
						(evalMD != null && evalMD.getLimitedRangeStartAt() > 0),
						(evalMD != null && evalMD.getLimitedRangeSize() < Integer.MAX_VALUE));
				if (debug) {
					log.log(BasicLevel.DEBUG, "SQL query for leaf " + leaf + ": " + sqlRequest);
				}
			} catch (Exception e) {
				throw new EvaluationException(e);
			}
		}
        try {
            ps = conn.prepareStatement(
                    sqlRequest,
                    ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_READ_ONLY);
            if (usedParameters != null) {
                assignParameter(ps, usedParameters, rdbAdapt);
            }
        } catch (Exception e) {
            throw new EvaluationException(e);
        }
		if (debug) {
			log.log(BasicLevel.DEBUG, "Evaluate the query: " + sqlRequest);
		}
		try {
			resultSet = ps.executeQuery();
			if (!resultSet.next()) {
				if (debug) log.log(BasicLevel.DEBUG, "No result, closing the ResultSet and the PreparedStatement");
				resultSet.close();
				ps.close();
				return new EmptyTupleCollection(lf.getTupleStructure());
			} else {
				if (debug) log.log(BasicLevel.DEBUG, "There is a result");
				return new JDBCTupleCollection(
						lf.getTupleStructure(), resultSet, ps, rdbAdapt, TClog);
			}
		} catch (EvaluationException ee) {
			throw ee;
		} catch (Exception mexp) {
			throw new EvaluationException(mexp);
		}
	}

	private Connection getConnection(Object connection) throws EvaluationException {
		try {
			return RdbConnectionWrapper.narrow2SQL(connection);
		} catch (PException e) {
			throw new EvaluationException(
					"Impossible to evaluate a jdbc leaf:", e);
		}
	}

	private RdbAdapter getRdbAdapter(RdbQueryLeaf leaf) {
		RdbAdapter rdbAdapt = null;
		if (leaf instanceof RdbExpQueryLeaf) {
			rdbAdapt = ((RdbExpQueryLeaf) leaf).getRdbAdapter();
		} else if (leaf instanceof AggregateRdbQueryNode) {
		    QueryTree[] children = ((AggregateRdbQueryNode) leaf).getChildren();
		    int i=0;
		    while(rdbAdapt == null) {
		        rdbAdapt = getRdbAdapter((RdbQueryLeaf) children[i]);
		        i++;
			}
		}
		if (rdbAdapt == null) {
			rdbAdapt = new BasicRdbAdapter();
		}
		return rdbAdapt;
	}

	private void assignParameter(PreparedStatement ps,
								 ParameterOperand[] usedParameters,
								 RdbAdapter rdbAdapt)
			throws MedorException, ExpressionException, SQLException, IOException {
	    int index = 1;
		for (int i = 0; i < usedParameters.length; i++) {
			ParameterOperand po = usedParameters[i];
			if (debug) {
				log.log(BasicLevel.DEBUG, "Assign parameter (name="
						+ po.getName() + ", value=" + po + ")");
			}
			switch (po.getType().getTypeCode()) {
			case QTypeTuple.TYPECODE_INT:
				rdbAdapt.setInt(ps, index, po.getInt());
				break;
			case QTypeTuple.TYPECODE_OBJINT:
				rdbAdapt.setOint(ps, index, (Integer) po.getObject());
				break;
			case QTypeTuple.TYPECODE_SHORT:
				rdbAdapt.setShort(ps, index, po.getShort());
				break;
			case QTypeTuple.TYPECODE_OBJSHORT:
				rdbAdapt.setOshort(ps, index, (Short) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BYTE:
				rdbAdapt.setByte(ps, index, po.getByte());
				break;
			case QTypeTuple.TYPECODE_OBJBYTE:
				rdbAdapt.setObyte(ps, index, (Byte) po.getObject());
				break;
			case QTypeTuple.TYPECODE_LONG:
				rdbAdapt.setLong(ps, index, po.getLong());
				break;
			case QTypeTuple.TYPECODE_OBJLONG:
				rdbAdapt.setOlong(ps, index, (Long) po.getObject());
				break;
			case QTypeTuple.TYPECODE_DOUBLE:
				rdbAdapt.setDouble(ps, index, po.getDouble());
				break;
			case QTypeTuple.TYPECODE_OBJDOUBLE:
				rdbAdapt.setOdouble(ps, index, (Double) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BOOLEAN:
				rdbAdapt.setBoolean(ps, index, po.getBoolean());
				break;
			case QTypeTuple.TYPECODE_OBJBOOLEAN:
				rdbAdapt.setOboolean(ps, index, (Boolean) po.getObject());
				break;
			case QTypeTuple.TYPECODE_FLOAT:
				rdbAdapt.setFloat(ps, index, po.getFloat());
				break;
			case QTypeTuple.TYPECODE_OBJFLOAT:
				rdbAdapt.setOfloat(ps, index, (Float) po.getObject());
				break;
			case QTypeTuple.TYPECODE_BIGDECIMAL:
			case QTypeTuple.TYPECODE_BIGINTEGER:
				rdbAdapt.setBigDecimal(ps, index, (BigDecimal) po.getObject());
				break;
			case QTypeTuple.TYPECODE_STRING:
				rdbAdapt.setString(ps, index, (String) po.getObject());
				break;
			case QTypeTuple.TYPECODE_CHAR:
				rdbAdapt.setChar(ps, index, po.getChar());
				break;
			case QTypeTuple.TYPECODE_OBJCHAR:
				rdbAdapt.setOchar(ps, index, (Character) po.getObject());
				break;
			case QTypeTuple.TYPECODE_SERIALIZED:
				rdbAdapt.setSerialized(ps, index, (Serializable) po.getObject());
				break;
			case QTypeTuple.TYPECODE_DATE:
				rdbAdapt.setDate(ps, index, (Date) po.getObject(), null);
				break;
			case QTypeTuple.TYPECODE_BYTEARRAY:
				rdbAdapt.setByteArray(ps, index, (byte[]) po.getObject());
				break;
			case QTypeTuple.TYPECODE_CHARARRAY:
				rdbAdapt.setCharArray(ps, index, (char[]) po.getObject());
				break;
			case QTypeTuple.TYPECODE_COLLECTION:
				//Already set in the sql request building
			    // It is not a parameter ==> keep the same index
			    index --;
				break;
			default:
				throw new EvaluationException("The type "
						+ po.getType().getJormName()
						+ " is not supported '"
						+ po.getType().getTypeCode() + "'"
				);
			}
			index ++;
		}
	}

	private ArrayList getUsedParameters(Expression e,
										ParameterOperand[] parameters,
										ArrayList al, 
										FilteredQueryTree fqt)
			throws EvaluationException {
	    if (e instanceof IsEmpty) {
	        Expression op = ((IsEmpty) e).getExpression(0);
	        if (!(op instanceof FieldOperand)) {
	            throw new IllegalStateException("can't handle operand type: " + op);
	        }
            QueryTree qt = ((QueryTreeField) ((FieldOperand) op).getField()).getQueryTree();
            if (!(qt instanceof FilteredQueryTree)) {
	            throw new IllegalStateException("can't querytree type: " + qt);
            }
            FilteredQueryTree _fqt = (FilteredQueryTree) qt;
            if (fqt != _fqt) {
                al = getUsedParameters(_fqt.getQueryFilter(), parameters, al, _fqt);
            }
/*	    } else if (e instanceof MemberOf) {
	        MemberOf mo = (MemberOf) e;
	        int moSize = mo.getOperandNumber() / 2;
	        //iterate on all pairs of Fields
	        for (int f = 0; f < moSize; f++) {
	            RdbExpField rdbField = (RdbExpField) ((FieldOperand) mo
	                    .getExpression(f + moSize)).getField();
	        }
			//TODO take memberof into account for parameters*/
	    } else if ((e instanceof InCollection) &&
				!(((Operator) e).getExpression(1) instanceof ParameterOperand)
		) {
			al = getUsedParameters(((Operator) e).getExpression(0),
					parameters, al, fqt);
			al = getUsedParametersInCollection(
					(Collection) ((Operand) ((Operator) e).getExpression(1)).getObject(),
					parameters,
					al);
		} else if (e instanceof Operator) {
            if (e instanceof Equal || e instanceof NotEqual) {
                Operator op = (Operator) e;
                Map ctx = Collections.singletonMap(
                        MedorExpression2WhereClause.POS, parameters);
                boolean nullValue_0;
                boolean nullValue_1;
                try {
                    nullValue_0 = MedorExpression2WhereClause.isNullParameterOperand(
                            op.getExpression(0), ctx);
                    nullValue_1 = MedorExpression2WhereClause.isNullParameterOperand(
                                op.getExpression(1), ctx);
                } catch (ExpressionException e1) {
                    throw new EvaluationException(e1);
                }
                if (nullValue_0 ^ nullValue_1) {
                    //Ignore the parameter because it has been replaced by 
                    // IS NULL or IS NOT NULL
                    return al;
                } 
            }
            for (int i = 0; i < ((Operator) e).getOperandNumber(); i++) {
				al = getUsedParameters(((Operator) e).getExpression(i),
						parameters, al, fqt);
			}
		} else if (e instanceof CalculatedParameterOperand) {
			CalculatedParameterOperand cpo = (CalculatedParameterOperand) e;
			try {
				cpo.evaluate(parameters);
			} catch (ExpressionException e1) {
				throw new EvaluationException(e1);
			}
			if (al == null)
				al = new ArrayList();
			al.add(cpo);
		} else if (e instanceof ParameterOperand) {
			int i = 0;
			ParameterOperand param = (ParameterOperand) e;
			while (i < parameters.length
					&& !parameters[i].getName().equals(param.getName()))
				i++;
			if (i < parameters.length) {
				if (parameters[i].getType().getTypeCode() !=
						param.getType().getTypeCode()) {
					throw new EvaluationException("Bad parameter type: found "
							+ parameters[i].getType().getJormName()
							+ " / expected " + param.getType().getJormName());
				}
				if (al == null)
					al = new ArrayList();
				al.add(parameters[i]);
			} else {
				StringBuffer msg = new StringBuffer();
				msg.append("You must specify the value of parameter ");
				msg.append(param.getName());
				msg.append(". Found: {");
				String sep = "";
				for (int j = 0; j < parameters.length; j++) {
					msg.append(sep);
					msg.append(parameters[j].getName());
					sep = ", ";
				}
				msg.append("}");
				throw new EvaluationException(msg.toString());
			}
		}
		return al;

	}

	private ArrayList getUsedParametersInCollection(
			Collection col,
			ParameterOperand[] parameters,
			ArrayList al)
			throws EvaluationException {
		Iterator colit = col.iterator();
		while (colit.hasNext()) {
			Object e = colit.next();
			if (e instanceof ParameterOperand) {
				int i = 0;
				ParameterOperand param = (ParameterOperand) e;
				while (i < parameters.length
						&& !parameters[i].getName().equals(param.getName()))
					i++;
				if (i < parameters.length) {
					if (parameters[i].getType().getTypeCode() !=
							param.getType().getTypeCode()) {
						throw new EvaluationException("Bad parameter type: found "
								+ parameters[i].getType().getJormName()
								+ " / expected " + param.getType().getJormName());
					}
					if (al == null)
						al = new ArrayList();
					al.add(parameters[i]);
				} else {
					String msg = "You must specify the value of parameter "
							+ param.getName() + ". Found: {";
					String sep = "";
					for (int j = 0; j < parameters.length; j++) {
						msg += sep + parameters[j].getName();
						sep = ", ";
					}
					msg += "}";
					throw new EvaluationException(msg);
				}
			}
		}
		return al;
	}

    private ParameterOperand[] getUsedPOs(QueryLeaf leaf, 
            ParameterOperand[] parameters,
            RdbAdapter rdbAdapt,
            EvaluationMetaData evalMD) throws EvaluationException {
        ArrayList usedParametersList = null;
        if (parameters != null
                && parameters.length > 0
                && leaf instanceof FilteredQueryTree) {
            Expression filter = ((FilteredQueryTree) leaf).getQueryFilter();
            if (filter != null) {
                usedParametersList = getUsedParameters(
                        filter, parameters, null, (FilteredQueryTree) leaf);
            }
        }
        //add the range parameters if necessary
        usedParametersList = addRangeParameters(usedParametersList, rdbAdapt, evalMD);
        if (usedParametersList != null && usedParametersList.size() > 0) {
            if (debug) {
                log.log(BasicLevel.DEBUG, "Used parameters: " + usedParametersList);
            }
            return (ParameterOperand[]) 
                usedParametersList.toArray(
                        new ParameterOperand[usedParametersList.size()]);
        }
        return null;
    }
    
	private String posToString(ParameterOperand[] pos) {
		if (pos == null) {
			return "null";
		}
		StringBuffer sb = new StringBuffer("[");
		String sep = "";
		for (int i = 0; i < pos.length; i++) {
			sb.append(sep);
			sb.append("(name='");
			sb.append("'");
			sb.append(pos[i].getName());
			sb.append("', value=");
			sb.append(pos[i]);
			sb.append(")");
		}
		sb.append("]");
		return sb.toString();
	}
	
	private ArrayList addRangeParameters(ArrayList al, 
	        RdbAdapter rdbAdapt, 
	        EvaluationMetaData evalMD) {
		if (evalMD != null) {
		    int rangePosition = rdbAdapt.getRangeParametersAtStart();
		    if (rangePosition != RdbAdapter.RANGEUNSUPPORTED) {
		        if (al == null) {
		            al = new ArrayList();
		        }
		        int rangeStart = evalMD.getLimitedRangeStartAt();
		        if (rangeStart > 0) {
		            BasicParameterOperand rangeStartParam = 
		                new BasicParameterOperand(PTypeSpace.INT,
		                    "rangeStart",
		                    rangeStart);
		            //position the parameter at the right place
		            if (rangePosition == RdbAdapter.RANGEATSTART) {
		                al.add(0, rangeStartParam);
		            } else {
		                al.add(rangeStartParam);
		            }
		        }
		        
		        int rangeSize = evalMD.getLimitedRangeSize();
		        if (rangeSize < Integer.MAX_VALUE) {
		            ParameterOperand rangeSizeParam = 
		                new BasicParameterOperand(PTypeSpace.INT,
		                    "rangeSize",
		                    rangeSize);
		            if (rangePosition == RdbAdapter.RANGEATSTART) {
		               if (rangeStart > 0) {
		                   al.add(1, rangeSizeParam);
		               } else { //case where start = 0, already one param added
		                   al.add(0, rangeSizeParam);
		               }
		            } else {
		                al.add(rangeSizeParam);
		            }
		            if (rangeStart > 0 && rdbAdapt instanceof OracleAdapter) {
		                al.add(new BasicParameterOperand(PTypeSpace.INT,
			                    "rangeStart",
			                    rangeStart));
		            }
		        }
		    }
		}
		return al;
	}

    public long deleteData(QueryLeaf ql, 
            ParameterOperand[] parameters, 
            Object connection, 
            EvaluationMetaData evalMD) throws EvaluationException {
        debug = log.isLoggable(BasicLevel.DEBUG);
        if (debug) {
            log.log(BasicLevel.DEBUG, "Parameters: " + posToString(parameters));
        }
        RdbQueryLeaf leaf = (RdbQueryLeaf) ql;
        //Calculate the parameter list of the leaf with the parameter list of the query
        RdbAdapter rdbAdapt = getRdbAdapter(leaf);
        Connection conn = getConnection(connection);
        ParameterOperand[] usedParameters = getUsedPOs(leaf, parameters, rdbAdapt, evalMD);
        String sqlDelete;
        synchronized (leaf) {
            try {
                sqlDelete = leaf.getSqlDelete(parameters);
            } catch (Exception e) {
                throw new EvaluationException(e);
            }
        }
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sqlDelete);
            if (usedParameters != null) {
                assignParameter(ps, usedParameters, rdbAdapt);
            }
        } catch (Exception e) {
            throw new EvaluationException(e);
        }
        if (debug) {
            log.log(BasicLevel.DEBUG, "Evaluate the query: " + sqlDelete);
        }
        int res = 0;
        try {
            res = ps.executeUpdate();
        } catch (Exception mexp) {
            throw new EvaluationException(mexp);
        } finally {
            try {
                ps.close();
            } catch (SQLException e) {
                log.log(BasicLevel.WARN, "impossible to close the PreparedStatement", e);
            }
        }
        return res;
    }

    public long updateData(QueryLeaf ql, 
            ParameterOperand[] parameters, 
            Object connection, 
            EvaluationMetaData evalMD) throws EvaluationException {
        debug = log.isLoggable(BasicLevel.DEBUG);
        if (debug) {
            log.log(BasicLevel.DEBUG, "Parameters: " + posToString(parameters));
        }
        RdbQueryLeaf leaf = (RdbQueryLeaf) ql;
        //Calculate the parameter list of the leaf with the parameter list of the query
        RdbAdapter rdbAdapt = getRdbAdapter(leaf);
        Connection conn = getConnection(connection);
        ParameterOperand[] usedParameters = getUsedPOs(leaf, parameters, rdbAdapt, evalMD);
        Expression updateExpression = (Expression) evalMD.getAnnotations()
            .get(EvaluationMetaData.ANNOTATION_UPDATE_EXPRESSION);
        if (updateExpression == null) {
            throw new EvaluationException("Impossible to perform an update without update Expression (no annotation found)");
        }
        PreparedStatement ps = null;
        String sqlUpdate;
        synchronized (leaf) {
            try {
                sqlUpdate = leaf.getSqlUpdate(parameters, updateExpression);
            } catch (Exception e) {
                throw new EvaluationException(e);
            }
        }
        try {
            ps = conn.prepareStatement(sqlUpdate);
            if (usedParameters != null) {
                assignParameter(ps, usedParameters, rdbAdapt);
            }
        } catch (Exception e) {
            throw new EvaluationException(e);
        }
        if (debug) {
            log.log(BasicLevel.DEBUG, "Evaluate the query: " + sqlUpdate);
        }
        int res = 0;
        try {
            res = ps.executeUpdate();
        } catch (Exception mexp) {
            throw new EvaluationException(mexp);
        } finally {
            try {
                ps.close();
            } catch (SQLException e) {
                log.log(BasicLevel.WARN, "impossible to close the PreparedStatement", e);
            }
        }
        return res;
    }
    
}
