/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

/**
 * Package definition.
 */
package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;


/**
 * This class is the implementation of a catesian result betwin two TupleCollection. Sub NodeEvaulators allow fetching
 * data from sub QueryTree.
 */
public class CartesianEvaluatedTC
		extends BasicBinaryEvalutedTC
		implements BinaryEvaluatedTC {

	private QueryNode query;
    private TupleCollection leftTC,rightTC;
    private ParameterOperand[] parameters;
    private Tuple currentBuffer;
    private Tuple leftTuple, rightTuple;
    private VariableOperand[] operandBuffers;
    private int cursor,size,leftTCSize,rightTCSize;

    public CartesianEvaluatedTC(QueryNode query, NodeEvaluator leftNodeEvaluator, NodeEvaluator rightNodeEvaluator, ParameterOperand[] parameters) throws MedorException {
        this.parameters = parameters;

        //get sizes
        leftTCSize = leftNodeEvaluator.getQueryNode().getTupleStructure().getSize();
        rightTCSize = rightNodeEvaluator.getQueryNode().getTupleStructure().getSize();
        size = query.getTupleStructure().getSize();

        //Init Buffers
        operandBuffers = new BasicVariableOperand[size];
        currentBuffer = new MemoryTuple(operandBuffers);

        // Fetching left and right TupleCollection Result
        rightTC = rightNodeEvaluator.fetchData(parameters);
        leftTC = leftNodeEvaluator.fetchData(parameters);
        if (isEmpty()) {
            cursor = -2;
        } else {
            init();
        }
    }

	public void close() throws MedorException {
		super.close();
		if (leftTC != null) {
			leftTC.close();
		}
		if (rightTC != null) {
			rightTC.close();
		}
	}

    public TupleStructure getMetaData() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return query.getTupleStructure();
    }

    public boolean isLast() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return ((rightTC.isLast()) && (leftTC.isLast()));
    }

    // Getting the first Result if possible
    private void init() throws MedorException {
        leftTC.first();
        rightTC.first();
        leftTuple = leftTC.getTuple();
        rightTuple = rightTC.getTuple();
        fixeTheLeftResult();
        fixeTheRightResult();
        cursor = 1;
    }

    public boolean next() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        boolean moved = false;
        if (isLast()) {
            moved = false;
        } else {
            moved = true;
            cursor++;
            if (rightTC.next()) {
                leftTuple = leftTC.getTuple();
                rightTuple = rightTC.getTuple();
                fixeTheRightResult();
            } else {
                if (leftTC.next()) {
                    rightTC.first();
                    rightTuple = rightTC.getTuple();
                    leftTuple = leftTC.getTuple();
                    fixeTheLeftResult();
                    fixeTheRightResult();
                }
            }
        }
        return moved;
    }

    public void first() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        //The Initialization
        init();
    }

    public int getRow() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return cursor;
    }

    public Tuple getTuple() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (!isEmpty() && (getRow() >= 1))
            return currentBuffer;
        else
            throw new MedorException(" No elements fetched in this TupleCollection  " + getRow());
    }

    public synchronized Tuple getTuple(int numTuple) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        //System.out.println(" l'index =  " + row(numTuple));
        return getTuple();
    }

    public boolean isEmpty() throws MedorException{
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return ((leftTC.isEmpty()) && (rightTC.isEmpty()));
    }

    public boolean row(int numTuple) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        first();
        int cpt = 1;
        boolean go = true;
        while ((cpt <= numTuple) && (cpt > 1) && go) {
            if (!next()) go = false;
        }
        if (!go)
            return false;
        else {
            cursor = numTuple;
            return true;
        }

    }

    private void fixeTheLeftResult() throws MedorException {
        for (int cpt = 0; (cpt < leftTCSize); cpt++) {
            operandBuffers[cpt] = (VariableOperand)leftTuple.getLikeOperand(cpt + 1);
        }
    }

    private void fixeTheRightResult() throws MedorException {
        for (int cpt = leftTCSize; (cpt < size); cpt++) {
            operandBuffers[cpt] = (VariableOperand)rightTuple.getLikeOperand(cpt - leftTCSize + 1);
        }
    }

    public int getLeftTCCursor() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return leftTC.getRow();
    }

    public int getRightTCCursor() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return rightTC.getRow();
    }
}
