/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

/**
 * Package definition.
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.api.EvaluationMetaData;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.filter.api.ExpressionHelper;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.postfix.PostfixExpressionHelper;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;

import java.util.LinkedList;

/**
 * This class represent is TupleCollection Result of the Nest Operation.
 * for each GroupBy fields, a SelectProject Operation is evaluated. The result
 * of this Operation is the NestedField TupleCollection. To select the next Group,
 * a collection containing indexes of all tuples already grouped.
 */
public class NestEvaluatedTC
		extends BasicBinaryEvalutedTC
		implements BinaryEvaluatedTC {

	private NestQueryNode nestedQueryTree;
	private QueryNode query1, query2;
	private Expression groupByFilter1,groupByFilter2;
	private int cursor,tcResultSize;
	private boolean isTheLast,nextEvaluatorIsFree;
	/**
	 * Contain indexes of all tuples already visited in the sub-TupleCollection
	*/
 	private LinkedList visitedGroup;
	private ExpressionHelper expressionHeleper;
	private NodeEvaluator subNodeEvaluator,nodeEvaluator1,nodeEvaluator2;
	/**
	 * Buffers containing current and next result Tuple
	 */
	private MemoryTuple tupleBuffer1,tupleBuffer2;
	/**
	 * Arrays of buffer's fields
 	 */
	private VariableOperand[] operandBuffer1,operandBuffer2;
	private TupleCollection nestedTupleCollection;
	/**
	 * Contains the attributes which are not grouped designed with their new Names
 	 */
	private Field[] groupByFields;
	private EvaluationMetaData evaluationMetaData;

	public NestEvaluatedTC(NestQueryNode query,
						   NodeEvaluator subNodeEvaluator,
						   ParameterOperand[] parameters,
						   EvaluationMetaData evaluationMetaData)
		throws MedorException {
		this.nestedQueryTree = query;
		this.subNodeEvaluator = subNodeEvaluator;
		this.evaluationMetaData = evaluationMetaData;
		visitedGroup = new LinkedList();
		nestedTupleCollection = subNodeEvaluator.fetchData(parameters);
		groupByFields = query.getNestingFields();
		tcResultSize = groupByFields.length;

		/**
		 * We construct a Filter wich is used to group Fields using the
		 * SelectProject Operation (Equality between same fields).
		 * We can use a parametered Filter (ParameterOperand) or FieldOperand.
		 * We have used the second method by upgrading the Filter Operand Values
		 * for each group.
		 */
		int sizeOfFilter = ((groupByFields.length - 1) * 4) + 3;
		Expression[] filterExpression = new Expression[sizeOfFilter];
		Expression[] filterExpressionClone = new Expression[sizeOfFilter];
		FieldOperand fop;

		int filterIndex = 0;
		for (int cpt = 0; (cpt < groupByFields.length); cpt++) {
			fop = new BasicFieldOperand(groupByFields[cpt]);
			if (filterIndex == 0) {
				filterExpression[0] = fop;
				filterExpressionClone[0] = fop;
				filterExpression[2] = new Equal();
				filterExpressionClone[2] = new Equal();
				filterIndex = 2;
			}
			else {
				filterExpression[++filterIndex] = fop;
				filterExpressionClone[++filterIndex] = fop;
				filterIndex = filterIndex + 2;
				filterExpression[filterIndex] = new Equal();
				filterExpressionClone[filterIndex] = new Equal();
				filterExpression[++filterIndex] = new And();
				filterExpressionClone[++filterIndex] = new And();
			}
		}
		expressionHeleper = new PostfixExpressionHelper();

		try {
			groupByFilter1 = expressionHeleper.join(filterExpression);
			groupByFilter2 = expressionHeleper.join(filterExpressionClone);
		}
		catch (ExpressionException e) {
			throw new MedorException(e);
		}

		operandBuffer1 = new VariableOperand[tcResultSize];
		operandBuffer2 = new VariableOperand[tcResultSize];
		for (int cpt = 0;
			 (cpt < (nestedQueryTree.getTupleStructure().getSize()));
			 cpt++) {

			// init the buffer Tuple
			operandBuffer1[cpt] =
				new BasicVariableOperand(
					nestedQueryTree.getTupleStructure().getField(cpt + 1).getType()
				);
			operandBuffer2[cpt] = new BasicVariableOperand(
				nestedQueryTree.getTupleStructure().getField(cpt + 1).getType()
			);
		}

		tupleBuffer1 = new MemoryTuple(operandBuffer1);// Buffer Tuple
		tupleBuffer2 = new MemoryTuple(operandBuffer2); // next Buffer Tuple
		init();
	}

	public void close() throws MedorException {
		super.close();
		if (nestedTupleCollection != null)
			nestedTupleCollection.close();
	}

	/**
	 * This method will change the SelectProject Filter in order to get the next
	 * GroupBy fields. The new values of the BasicFieldOperand are modified
	 * corresponding to the next field values of the next GroupBy Fields values.
	 */
	private void changeGroupByFilter(Expression filter, Operand[] values)
		throws ExpressionException {
		Expression[] exp = expressionHeleper.toStack(filter);
		exp[1] = values[0];
		int rg = 4;
		for (int i = 1; (i < values.length - 1); i++) {
			exp[rg] = values[i];
			rg = rg + 4;
		}
	}

	/**
	 * This function will move the cursor of the sub TupleCollection at the row
	 * of the next GroupBy fileds and load it.
	 * In oure case, we iterate over all Tuples of the the sub TupleCollection
	 * and surch verify if the current Tuple exists or not in the visitedGroup
	 * collection.
	 * @param dest is the buffer Tuple
	 * @return true if there is next GroupBy and false if not
	 * @throws MedorException if subTupleCollection iteration error
	 */
	private boolean prepareNextNotVisitedGroupIn(VariableOperand[] dest)
		throws MedorException, ExpressionException {
		Tuple tuple = null;
		boolean stop = false;
		while (!(stop) && nestedTupleCollection.next()) {
			if (!(visitedGroup.contains(
				new Integer(nestedTupleCollection.getRow())
			))) {
				visitedGroup.add(new Integer(nestedTupleCollection.getRow()));
				stop = true;
				tuple = nestedTupleCollection.getTuple();
			}
		}
		if (stop) {
			// load the next GroupBy fields
			nestedQueryTree.getTupleLoader().loadTuple(tuple, dest, null);
			return true;
		}
		else
			return false;
	}

	/**
	 * This method will get the two first Tuples if possible and tell
	 * if it is the last tuple or not;
	 */
	private void init() throws MedorException {
		cursor = 1; //Positionning at the first Tuple
		nextEvaluatorIsFree = false;

		try {
			//We prepare the SelectProject Operation wich will represents the
			//result of the Nested Field
			nestedQueryTree.getTupleLoader().loadTuple(
				nestedTupleCollection.getTuple(), operandBuffer1, null);
			visitedGroup.add(new Integer(1)); // The first GroupBy Tuple index
			changeGroupByFilter(groupByFilter1, operandBuffer1);

			query1 = new SelectProject(groupByFilter1, "");
			//add the projected fields which are the nested fields
			Field[] nestedFields1 =
				(nestedQueryTree.getNestedField()).getFields();
			for (int i1 = 0; i1 < nestedFields1.length; i1++) {
				query1.addPropagatedField(nestedFields1[i1].getName(),
					nestedFields1[i1].getType(),
					new QueryTreeField[]{(QueryTreeField) nestedFields1[i1]});
			}
			nodeEvaluator1 =
				new UnaryIteratifNodeEvaluator(query1,
						subNodeEvaluator, evaluationMetaData);

			// the nestedField TupleCollection Result
			TupleCollection tctemp = nodeEvaluator1.fetchData(null);
			BinaryEvaluatedTC itc = (BinaryEvaluatedTC) tctemp;

			// we mark or the tuples already grouped
			do {
				visitedGroup.add(new Integer(itc.getRightTCCursor()));
			} while (tctemp.next());

			tctemp.first();
			// Must be changed for more object management
			operandBuffer1[tcResultSize - 1].setValue(tctemp);

			//Now we prepare the next group if possible
			if (!(prepareNextNotVisitedGroupIn(operandBuffer2))) {
				isTheLast = true;
			}
			else {
				changeGroupByFilter(groupByFilter2, operandBuffer2);
				query2 = new SelectProject(groupByFilter1, "");
				//add the projected fields which are the nested fields
				Field[] nestedFields2 =
					(nestedQueryTree.getNestedField()).getFields();
				for (int i2 = 0; i2 < nestedFields2.length; i2++) {
					query2.addPropagatedField(nestedFields2[i2].getName(),
						nestedFields2[i2].getType(),
						new QueryTreeField[]{(QueryTreeField) nestedFields2[i2]});
				}
				nodeEvaluator2 =
					new UnaryIteratifNodeEvaluator(query1,
							subNodeEvaluator, evaluationMetaData);
				TupleCollection tctemp1 = nodeEvaluator2.fetchData(null);
				BinaryEvaluatedTC itc1 = (BinaryEvaluatedTC) tctemp1;
				visitedGroup.add(new Integer(nestedTupleCollection.getRow()));

				do {
					visitedGroup.add(new Integer(itc1.getRightTCCursor()));
				} while (tctemp1.next());
				tctemp1.first();

				// Affecting the NestedField TupleCollection Result
				operandBuffer2[tcResultSize - 1].setValue(tctemp1);
			}
		}
		catch (MedorException e) {
			throw e;
		}
		catch (ExpressionException e) {
			throw new MedorException(e);
		}
	}

	public TupleStructure getMetaData() throws MedorException {
		return nestedQueryTree.getTupleStructure();
	}

	public boolean isLast() throws MedorException {
		return isTheLast;
	}

	public boolean next() throws MedorException {
		boolean deplacer = false;
		if (isEmpty() || (cursor == -2))
			deplacer = false;
		else if (isLast()) {
			// we have already visited the last Tuple
			tupleBuffer1 = null;
			tupleBuffer2 = null;
			cursor = -2;
			deplacer = false;
		}
		else {
			deplacer = true;
			// we are effectilly passed to the next Tuple.
			// We verify if it is the last one
			cursor++;
			BinaryEvaluatedTC itc = null;
			TupleCollection tctemp = null;
			try {
				if (nextEvaluatorIsFree) {
					// This means that i can use the seconde evaluator
					if (!(prepareNextNotVisitedGroupIn(operandBuffer2))) {
						isTheLast = true;
						nextEvaluatorIsFree = false;
					}
					else {
						changeGroupByFilter(groupByFilter2, operandBuffer2);
						tctemp = nodeEvaluator2.fetchData(null);
						itc = (BinaryEvaluatedTC) tctemp;
						visitedGroup.add(
							new Integer(nestedTupleCollection.getRow())
						);
						do {
							visitedGroup.add(new Integer(itc.getRightTCCursor()));
						} while (tctemp.next());
						tctemp.first();
						// The last Field is a TupleCollection
						operandBuffer2[tcResultSize - 1].setValue(tctemp);
						nextEvaluatorIsFree = false;
					}
				}
				else {
					if (!(prepareNextNotVisitedGroupIn(operandBuffer1))) {
						isTheLast = true;
						nextEvaluatorIsFree = true;
					}
					else {
						changeGroupByFilter(groupByFilter1, operandBuffer1);
						tctemp = nodeEvaluator1.fetchData(null);
						itc = (BinaryEvaluatedTC) tctemp;
						visitedGroup.add(
							new Integer(nestedTupleCollection.getRow())
						);
						do {
							visitedGroup.add(new Integer(itc.getRightTCCursor()));
						} while (tctemp.next());
						tctemp.first();
						// The last field is a TupleCollection;
						operandBuffer1[tcResultSize - 1].setValue(tctemp);
						nextEvaluatorIsFree = true;
					}
				}
			}
			catch (MedorException e) {
				throw e;
			}
			catch (ExpressionException e) {
				throw new MedorException(e);
			}
		}
		return deplacer;
	}

	public void first() throws MedorException {
		init();
	}

	public int getRow() throws MedorException {
		return cursor;
	}

	public boolean isEmpty() throws MedorException {
		return nestedTupleCollection.isEmpty();
	}

	public Tuple getTuple() throws MedorException {
		if (!isEmpty() && (getRow() >= 1)) {
			if (nextEvaluatorIsFree)
				return tupleBuffer2;
			else
				return tupleBuffer1;
		}
		else
			throw new MedorException(" No elements fetched in this TupleCollection  "
				+ getRow());
	}

	public synchronized Tuple getTuple(int numTuple) throws MedorException {
		return getTuple();
	}

	public TupleCollection getGroupByTupleCollection() throws MedorException {
		if ((!(isEmpty())) && (getRow() >= 1)) {
			TupleCollection otemp =
				tupleBuffer1.getTupleCollection(tcResultSize);
			TupleCollection tc = otemp;
			return tc;
		}
		else
			throw new MedorException(" No elements fetched in this TupleCollection  "
				+ getRow());
	}

	public boolean row(int numTuple) throws MedorException {
		first();
		int cpt = 1;
		boolean go = true;
		while ((cpt <= numTuple) && (cpt > 1) && go) {
			if (!next()) go = false;
		}
		if (!go)
			return false;
		else {
			cursor = numTuple;
			return true;
		}
	}

	public int getLeftTCCursor() throws MedorException {
		return this.nestedTupleCollection.getRow();
	}

	public int getRightTCCursor() throws MedorException {
		return -1;
	}
}
