/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferHolder;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * A project Operation QueryNode is performed by this TupleCollection. This
 * operation is not costs and doeas not uses buffers.
 */
public class ProjectEvaluatedTC
		extends BasicBinaryEvalutedTC
		implements BinaryEvaluatedTC, PrefetchBufferHolder {

	private QueryNode query;
	private TupleCollection subResultTC;
	private ParameterOperand[] parameters;
	private Tuple currentBuffer;
	private VariableOperand[] currentAttBuffers;
	private int cursor,size;
	private boolean empty;
	//boolean to indicate whether close() has been explicitely called
	private boolean userHasClosed = false;
	private PrefetchBuffer prefetchBuffer;

	public ProjectEvaluatedTC(QueryNode query,
							  NodeEvaluator subNodeEvaluator,
							  ParameterOperand[] parameters,
							  PrefetchBuffer pb) throws MedorException {
		super();
		this.parameters = parameters;
		this.query = query;
		this.prefetchBuffer = pb;
		if (prefetchBuffer != null) {
			pb.setTupleCollection(this);
		}
		// the size of the result of the projection operation
		size = query.getTupleStructure().getSize();
		currentAttBuffers = new BasicVariableOperand[size];
		for (int cpt = 0; (cpt < query.getTupleStructure().getSize()); cpt++) {
			currentAttBuffers[cpt] = new BasicVariableOperand(
					query.getTupleStructure().getField(cpt + 1).getType());
		}
		// This object is used as a buffer
		currentBuffer = new MemoryTuple(currentAttBuffers);
		subResultTC = subNodeEvaluator.fetchData(parameters);
		empty = (subResultTC.isEmpty());
		if (empty) {
			cursor = 0;
		} else {
			init();
		}
	}

	public void invalidatePrefetchBuffer() throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "Invalidating PrefetchBuffer " + prefetchBuffer + " for " + this);
		}
		prefetchBuffer = null;
		if (userHasClosed) {
			if (debug) {
				log.log(BasicLevel.DEBUG, "The tupleCollection was closed earlier: closing it for real");
			}
			this.close();
		}
	}

	public void close() throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "Closing TupleCollection " + this);
		}
		userHasClosed = true;
		//closing only if the prefetchBuffer is null or has been closed.
		if (prefetchBuffer == null || prefetchBuffer.isClosed()) {
			if (debug) {
				log.log(BasicLevel.DEBUG, "PrefetchBuffer " + prefetchBuffer + " is null or was closed previously: real close of the TupleCollection");
			}
			prefetchBuffer = null;
			super.close();
			if (subResultTC != null) {
				subResultTC.close();
			}
		}
	}

	public TupleStructure getMetaData() throws MedorException {
		return query.getTupleStructure();
	}

	public boolean isLast() throws MedorException {
		return subResultTC.isLast();
	}

	public int card() throws MedorException {
		return cursor;
	}

	public boolean next() throws MedorException {
		boolean moved;
		if (subResultTC.isLast()) {
			// we have already iterate the last tuple
			cursor = -2;
			moved = false;
		} else {
			cursor++;
			subResultTC.next();
			//Upgrade the buffer
			getNextResult();
			if (prefetchBuffer != null) {
				prefetchBuffer.addPrefetchTuple();
			}
			moved = true;
		}
		return moved;
	}

	private void init() throws MedorException {
		subResultTC.first();
		getNextResult();
		cursor = 1;
		if (prefetchBuffer != null) {
			prefetchBuffer.addPrefetchTuple();
		}
	}

	public void first() throws MedorException {
		//The Initialisation
		init();
	}

	public int getRow() throws MedorException {
		return cursor;
	}

	public Tuple getTuple() throws MedorException {
		if (!isEmpty() && (getRow() >= 1))
			return currentBuffer;
		else
			throw new MedorException(" No elements fetched in this TupleCollection  " + getRow());
	}

	public synchronized Tuple getTuple(int numTuple) throws MedorException {
		row(numTuple);
		return getTuple();
	}

	public boolean isEmpty() throws MedorException {
		return (subResultTC.isEmpty());
	}

	public boolean row(int numTuple) throws MedorException {
		boolean go = subResultTC.row(numTuple);
		if (!go) {
			return false;
		} else {
			getNextResult();
			cursor = numTuple;
			return true;
		}
	}

	public int getLeftTCCursor() throws MedorException {
		return subResultTC.getRow();
	}

	public int getRightTCCursor() throws MedorException {
		return -1;
	}

	private void getNextResult() throws MedorException {
		try {
			query.getTupleLoader().loadTuple(
					subResultTC.getTuple(), currentAttBuffers, parameters);
		}
		catch (ExpressionException e) {
			throw new MedorException(e);
		}
	}
}
