/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.eval.prefetch.api;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;

/**
 * Caches a set of prefetched tuples as long as they are valid within the
 * connection context within which they have been retrieved.
 * @author  P. Dechamboux
 */
public interface PrefetchCache {
    /**
     * Retrieves the prefetched tuple from this cache associated with this
     * index information, within the given caching context. The prefetched tuple
     * also belongs to the given object category.
     * @param index The index that allows the yielded prefetched data item to be
     *              identified into the cache.
     * @param cat   The category (e.g., class) to which the objects, that can be
     *              built using the prefetched data, belong to.
     * @param ctxt  The caching context within which the data has been
     *              prefetched.
     * @return      The prefetched tuple.
     */
    Tuple getPrefetchTuple(Object index, Object cat, Object ctxt)
            throws MedorException;

    /**
     *  Same as getPrefetchTuple, except it returns the TupleCollection.
     */
    TupleCollection getPrefetchTupleCollection(Object index, Object cat, Object ctxt)
    throws MedorException;
    
    /**
     * Invalidates all the prefetched tuples that are associated with the given
     * caching connection context.
     * @param ctxt  The invalid caching context.
     */
    void invalidatePrefetchBuffer(Object ctxt);

    /**
     * Allocates a prefetched buffer for the object category and the connection
     * context in order to register prefetched tuples.
	 * @param pbf		The factory for creating actual PrefetchBuffer
     * @param cat       The associated object category.
     * @param ctxt      The associated caching context.
     * @param indexpos  The position of the index field within the prefetched
     *                  tuples.
     * @param register if true, call the register method
     * @return  The allocated PrefetchBuffer.
     */
    PrefetchBuffer createPrefetchBuffer(PrefetchBufferFactory pbf,
										Object cat,
										Object ctxt,
										int indexpos,
										boolean register)
            throws MedorException;
    /**
     * Allocates a prefetched buffer for the object category and the connection
     * context in order to register prefetched tuples.
	 * @param pbf		The factory for creating actual PrefetchBuffer
     * @param cat       The associated object category.
     * @param ctxt      The associated caching context.
     * @param indexpos  The position of the index field within the prefetched
	 * @param multithread indicates if the context is used by several threads
	 * @param register if true, call the register method
     * @return  The allocated PrefetchBuffer.
     */
    PrefetchBuffer createPrefetchBuffer(PrefetchBufferFactory pbf,
										Object cat,
										Object ctxt,
										int indexpos,
										boolean multithread,
										boolean register)
            throws MedorException;
    
    /**
     * Register the prefetched buffer pb for the object category and the connection
     * context in order to register prefetched tuples.
	 * @param pbf		 The PrefetchBuffer to be registered for the object category
     * @param cat       The associated object category.
     * @param ctxt      The associated caching context.
     * @return true if the registration is done, false else
     */
    boolean registerPrefetchBuffer(PrefetchBuffer pb,
										Object cat,
										Object ctxt)
            throws MedorException;
    
    /**
     * clean the prefetch cache.
     *
     */
    void clean();
}
