/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.eval.prefetch.lib;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferFactory;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * @author  P. Dechamboux
 */
public class PrefetchCacheImpl implements PrefetchCache {
    private HashMap txContexts = new HashMap();

	Logger logger;

	public PrefetchCacheImpl() {
	}

	public PrefetchCacheImpl(Logger logger) {
		this.logger = logger;
	}

	public Logger getLogger() {
		return logger;
	}

	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	// IMPLEMENTATION OF METHODS FROM THE PrefetchCache INTERFACE

    public Tuple getPrefetchTuple(Object index, Object cat, Object ctxt)
            throws MedorException {
    	if (ctxt == null) {
            throw new MedorException("Cannot search a prefetch tuple without a transaction context!");
        }
        if (cat == null) {
            throw new MedorException("Cannot search a prefetch tuple without an object category!");
        }
        HashMap catmap = (HashMap) txContexts.get(ctxt);
        if (catmap == null) {
            return null;
        }
        ArrayList pblist = (ArrayList) catmap.get(cat);
        if (pblist == null) {
            return null;
        }
        Tuple res = null, tmp;
        for(int i=0; i<pblist.size(); i++) {
            PrefetchBuffer pb = (PrefetchBuffer) pblist.get(i);
            tmp = pb.getTuple(index);
            if (res == null && tmp != null) {
                res = tmp;
            }
        }
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "getPrefetchTupleCollection: "
					+ "\n\tcategory=" + cat
					+ "\n\tcontext=" + ctxt
					+ "\n\tindex=" + index
					+ "\n\t==>" + printTuple(res));
		}
        return res;	
    }

    public TupleCollection getPrefetchTupleCollection(Object index, Object cat, Object ctxt)
    throws MedorException {
    	if (ctxt == null) {
            throw new MedorException("Cannot search a prefetch tuple without a transaction context!");
        }
        if (cat == null) {
            throw new MedorException("Cannot search a prefetch tuple without an object category!");
        }
        HashMap catmap = (HashMap) txContexts.get(ctxt);
        if (catmap == null) {
            return null;
        }
        ArrayList pblist = (ArrayList) catmap.get(cat);
        if (pblist == null) {
            return null;
        }
        TupleCollection res = null, tmp;
        for(int i=0; i<pblist.size(); i++) {
            PrefetchBuffer pb = (PrefetchBuffer) pblist.get(i);
            tmp = pb.getTupleCollection(index);
            if (res == null && tmp != null) {
                res = tmp;
            }
        }
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "getPrefetchTupleCollection: "
					+ "\n\tcategory=" + cat
					+ "\n\tcontext=" + ctxt
					+ "\n\tindex=" + index
					+ "\n\t==>" + printTuple(res.getTuple()));
		}
        return res;	
    }
    
    public void invalidatePrefetchBuffer(Object ctxt) {
        HashMap catmap = (HashMap) txContexts.get(ctxt);
        if (catmap == null) {
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "Invalidation of an unregistered context: " + ctxt);
			}
            return;
        }
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Invalidation of the context: " + ctxt);
		}
        txContexts.remove(ctxt);
        Iterator it = catmap.values().iterator();
        while (it.hasNext()) {
            ArrayList pblist = (ArrayList) it.next();
            Iterator it2 = pblist.iterator();
            while (it2.hasNext()) {
				try {
					((PrefetchBuffer) it2.next()).close();
				} catch (MedorException e) {
				}
			}
        }
    }

    public PrefetchBuffer createPrefetchBuffer(PrefetchBufferFactory pbf, Object cat, Object ctxt, int indexpos, boolean register)
		throws MedorException {
		return createPrefetchBuffer(null, cat, ctxt, indexpos, false, register);
	}

    public PrefetchBuffer createPrefetchBuffer(PrefetchBufferFactory pbf, Object cat, Object ctxt, int indexpos, boolean multithread, boolean register)
            throws MedorException {
        PrefetchBuffer res = new PrefetchBufferImpl(indexpos, multithread, logger);
        if (register) {
        	registerPrefetchBuffer(res, cat, ctxt);
        }
        return res;
    }

    public boolean registerPrefetchBuffer(PrefetchBuffer pb, Object cat, Object ctxt)
    		throws MedorException {
    	if (ctxt == null) {
            return false;
        }
        if (cat == null) {
            throw new MedorException("Cannot register a prefetch buffer without an object category!");
        }
        HashMap catmap = (HashMap) txContexts.get(ctxt);
        if (catmap == null) {
			synchronized(txContexts) {
				catmap = (HashMap) txContexts.get(ctxt);
				if (catmap == null) {
					catmap = new HashMap();
					txContexts.put(ctxt, catmap);
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "create a map for the context");
					}

				}
			}
        }
        ArrayList pblist = (ArrayList) catmap.get(cat);
        if (pblist == null) {
			synchronized(catmap) {
				pblist = (ArrayList) catmap.get(cat);
				if (pblist == null) {
					pblist = new ArrayList();
					catmap.put(cat, pblist);
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "create a list of PrefetchBuffer");
					}
				}
			}
        }
        pblist.add(pb);
        if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "PrefetchBuffer created and added: " + pb);
        }
        return true;
    }
    
	public static String printTuple(Tuple t) {
		if (t == null) {
			return null;
		}
		StringBuffer sb = new StringBuffer();
		String sep = "(";
		int size = t.getSize();
		for(int i=1; i<=size; i++) {
			sb.append(sep);
			sep = ", ";
			sb.append(i);
			sb.append(": ");
			try {
				sb.append(t.getObject(i));
			} catch (MedorException e) {
				sb.append("not printable");
			}
		}
		return sb.toString();
	}
	
	public void clean() {
		txContexts.clear();
	}
}
