/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * A class loader that always checks a list of special directories on the hard
 * disk to load classes or Java source files when compilation if needed.
 * <p> The
 *  principal method loadClass(name,resolve) is coded so that name represents a
 * Class file name or a Java source file. In the second case, the Java source
 * file is compiled and loaded. The class loaded are used in the current
 * application, so loading the class referenced by these classes is delegated
 * to the class system loader if they are already loaded.
 */
public class DirsCompileClassLoader extends ClassLoader {

    private long bytecount;
    private Hashtable cache = new Hashtable();
    private HashSet pathList = new HashSet();

    public DirsCompileClassLoader(HashSet classpaths) {
        super();
        pathList = classpaths;
    }

    // load the class data from the class file in specified dir
    private byte[] loadClassData(String name) throws ClassNotFoundException {

        byte[] data = null;
        try {
            String path = getJavaPathClass(name);
            File cf = new File(path + File.separatorChar +
				name.replace('.', File.separatorChar) + ".class");
            FileInputStream in = new FileInputStream(cf);
            bytecount = cf.length();
            if (bytecount > 0) {
                data = new byte[(int) bytecount];
                in.read(data);
                in.close();
            }
        } catch (IOException ioexp) {
            throw new ClassNotFoundException(name);
        }
        return (data);
    }

    public synchronized Class loadClass(String name, boolean resolve)
		throws ClassNotFoundException {
        Class c = (Class) cache.get(name);
        System.out.println("Loading ..." + name);
        if (c == null) {
            // the class has not been loaded  yet
            // When class to load need others classes, for example: java classes, they are directly loaded by delegation to
            // the system class Loader. The directory must be empty to compile the java class.
            try {
                System.out.println("First try default System loading class... ");
                c = this.findSystemClass(name);// don't load classes already loaded by the application class loader
            } catch (ClassNotFoundException e) {
                System.err.println("System loading... failed when loading  " + name);
            }
            if (c == null) {
                // In this case the name represents a java source name, so we must compile at first
                if ((!isClass(name)) && (isJavaSource(name))) {
                    Runtime rt = Runtime.getRuntime();
                    Process p;
                    // Compilation by javac or we can imagine an other compilator
                    try {
                        System.out.println("Launching javac ..." + pathList);
                        String classpath = "";
                        Iterator i = pathList.iterator();
                        if (i.hasNext()) classpath =
							classpath + ((String) i.next());
                        while (i.hasNext()) {
							// Only For Windows classpath
                            classpath = classpath + ";" + ((String) i.next());
                        }

                        String cmd = "javac -classpath " + classpath + " " +
							getJavaSourcePath(name) + File.separatorChar +
							name.replace('.', File.separatorChar) + ".java";
                        System.out.println(cmd);
                        p = rt.exec(cmd);
                        p.waitFor();
                        System.out.println("Compilation finisched ...");

                    } catch (IOException e) {
                        System.err.println(" Compilation failed... " + e.getMessage());


                    } catch (InterruptedException e) {
                        System.err.println(" Compilation failed... " + e.getMessage());
                    }

                    c = loadClass(name, resolve);
                } else if (isClass(name)) {

                    System.err.println("Try loading class data bytes... " + name);
                    byte[] data = loadClassData(name);

                    if (data != null) {
                        System.out.println("Byte class loaded...try to define class now");
                        c = defineClass(name, data, 0, (int) bytecount);
                        if (resolve && c != null) {
                            resolveClass(c);
                        }
                    }
                } else
                    throw new ClassNotFoundException(name);
				// name is not boyh class or java source file
           }
        }


        if (c != null)
            cache.put(name, c);
        else
            throw new ClassNotFoundException(name);
        return c;
    }

    private boolean isClass(String name) {
        boolean isClass = false;
        Iterator it = pathList.iterator();
        String dir = null;
        while ((it.hasNext()) & (!isClass)) {
            dir = (String) it.next();
            //System.out.println(dir + File.separatorChar + name.replace('.', File.separatorChar) + ".class");
            File cf = new File(dir + File.separatorChar +
				name.replace('.', File.separatorChar) + ".class");
            if (cf.exists()) isClass = true;

        }
        return isClass;
    }

    private boolean isJavaSource(String name) {
        boolean isSource = false;
        Iterator it = pathList.iterator();
        String dir = null;
        while (it.hasNext() && (!isSource)) {
            dir = (String) it.next();
            //System.out.println(dir + File.separatorChar + name.replace('.', File.separatorChar) + ".java");
            File cf = new File(dir + File.separator +
				name.replace('.', File.separatorChar) + ".java");
            if (cf.exists()) isSource = true;

        }
        return isSource;
    }

    private String getJavaSourcePath(String name) {
        Iterator it = pathList.iterator();
        String dir = null;
        boolean trouv = false;
        while ((it.hasNext()) && (!trouv)) {
            dir = (String) it.next();
            // Second case: name represents a java source file
            File sf = new File(dir + File.separatorChar +
				name.replace('.', File.separatorChar) +
				".java");
            if (sf.exists()) trouv = true;
        }
        if (trouv)
            return dir;
        else
            return null;
    }

    private String getJavaPathClass(String name) {
        Iterator it = pathList.iterator();
        String dir = null;
        boolean trouv = false;
        while ((it.hasNext()) && (!trouv)) {
            dir = (String) it.next();
            // Second case: name represents a java source file
            File sf = new File(dir + File.separatorChar +
				name.replace('.', File.separatorChar) + ".class");
            if (sf.exists()) trouv = true;
        }
        if (trouv)
            return dir;
        else
            return null;
    }
}
