/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class represents the rule to drop projected fields which are not used.
 * <p>This rule is to be called on the root of the tree to be optimized.
 * The projected fields of the root node will not be deleted (they are
 * requested by the user).
 * <p>
 * As a remark, the user should be careful and not request unnecessary fields!
 */
public class DropUnusedProjFieldsRule extends BasicRule {

	public DropUnusedProjFieldsRule() {
		super("DropUnusedProjFieldsRule");
	}

	/**
	 * Relies on a Map which associates to each child the list of required
	 * Fields for that child.
	 */
	public QueryTree rewrite(QueryTree qt, QueryNode _parent)
			throws MedorException {
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		if (qt instanceof QueryNode) {
			Field[] myFields = qt.getTupleStructure().getFields();
			dropUnusedProjFields(qt, array2arrayList(myFields));
		}
		return qt;
	}

	private void dropUnusedProjFields(QueryTree qt,
									  ArrayList wanted)
		throws MedorException {
		if (debug) log.log(BasicLevel.DEBUG, "-Current node is " + qt);
		if (debug) log.log(BasicLevel.DEBUG, "Query tree before dropping fields:");
		if (debug) QueryTreePrinter.printQueryTree(qt, log);

		Map qt2ReqFields = new HashMap();

		if (debug) log.log(BasicLevel.DEBUG, " Drop unused fields from projected fields in according to usage");
		//drop unused fields from projected fields
		Field[] myFields = qt.getTupleStructure().getFields();
		for (int i = 0; i < myFields.length; i++) {
			if (debug) log.log(BasicLevel.DEBUG, "\tfield=" + myFields[i].getName());
			if (!wanted.contains(myFields[i])) {
				if (debug) log.log(BasicLevel.DEBUG, "\tDropping field " + myFields[i].getName() + " (" + myFields[i] + ") " + " qt=" + qt);
				if (qt instanceof QueryNode) {
					((QueryNode) qt).removeField(myFields[i].getName());
				}
				else if (qt instanceof RdbExpQueryLeaf)
					((RdbExpQueryLeaf) qt).removeRdbField(
						(RdbExpField) myFields[i]);
			}
		}

		if (debug) log.log(BasicLevel.DEBUG, "Query tree after dropping fields:");
		if (debug) QueryTreePrinter.printQueryTree(qt, log);

		if (qt instanceof QueryLeaf)
			return;

		//calculate the Map (qt child => { field }
		if (debug) log.log(BasicLevel.DEBUG, " calculate the Map (qt child => { field }");
		myFields = qt.getTupleStructure().getFields();
		for (int i = 0; i < myFields.length; i++) {
			if (debug) log.log(BasicLevel.DEBUG, "\tfield " + myFields[i].getName() + " is " + myFields[i]);
			if (myFields[i] instanceof CalculatedField)
				addFieldFromExpression(qt2ReqFields,
					((CalculatedField) myFields[i]).getExpression());
			else
				addFieldInMap(qt2ReqFields, (QueryTreeField) myFields[i]);
			//if other case, too bad
		}
		
		if (qt instanceof NestQueryNode) {
			if (debug) log.log(BasicLevel.DEBUG, " adding grouping fields");
			myFields = ((NestQueryNode) qt).getNestingFields();
			for (int i = 0; i < myFields.length; i++) {
			    
/*			    if (myFields[i] instanceof CalculatedField
				        && ((CalculatedField) myFields[i]).getExpression() instanceof SinglePName) {
				    log.log(BasicLevel.DEBUG, "Jorm2Rdb: found SinglePName");
				    QueryTreeField toadd = (QueryTreeField) 
				    ((SinglePName) ((CalculatedField) myFields[i]).getExpression()).getField().getField();
				    addAncestorsToMap(qt2ReqFields, new Field[] {(QueryTreeField) toadd});
			    } else {
			    */
			        addAncestorsToMap(qt2ReqFields, new Field[] {(QueryTreeField) myFields[i]});
/*			    }
 */

			}
		}

		if (debug) log.log(BasicLevel.DEBUG, "Add field from the query filter");
		addFieldFromExpression(qt2ReqFields,
			((FilteredQueryTree) qt).getQueryFilter());

		if (debug) log.log(BasicLevel.DEBUG, "Recall on children");
		//call on children
		for (Iterator it = qt2ReqFields.entrySet().iterator(); it.hasNext();) {
			Map.Entry e = (Map.Entry) it.next();
			dropUnusedProjFields((QueryTree) e.getKey(),
				(ArrayList) e.getValue());
		}
	}

	/**
	 * Adds the fields which are contained in a FieldOperand.
	 */
	private void addFieldFromExpression(Map qt2ReqFields, Expression exp) {
		if (exp instanceof FieldOperand) {
			QueryTreeField f = (QueryTreeField) ((FieldOperand) exp).getField();
			QueryTree qt = f.getQueryTree();
			ArrayList fields = (ArrayList) qt2ReqFields.get(qt);
			if (fields == null) {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldFromExpression: FieldOperand, new Qt in map");
				fields = new ArrayList(1);
				fields.add(f);
				qt2ReqFields.put(qt, fields);
			}
			else if (!(fields.contains(f))) {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldFromExpression: FieldOperand, Qt exists in map");
				fields.add(f);
			} else {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldFromExpression: FieldOperand already referenced");
			}
		} else if (exp instanceof Operator) {
			if (debug) log.log(BasicLevel.DEBUG, "addFieldFromExpression: Operator");
			for(int i=0; i<((Operator) exp).getOperandNumber(); i++) {
				addFieldFromExpression(qt2ReqFields,
					((Operator) exp).getExpression(i));
			}
		}
	}

	private void addFieldInMap(Map qt2ReqFields, QueryTreeField field) {
        if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: field " + field);
		Field[] anc = null;
		if (field instanceof PropagatedField) {
			anc = ((PropagatedField) field).getPreviousFields();
			if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: PropagatedField " + field.getName());
		}
		else if (field instanceof NestedField) {
			anc = ((NestedField) field).getFields();
			if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: NestedField " + field.getName());
		}
        else
            anc = new Field[0];
		addAncestorsToMap(qt2ReqFields, anc);
	}
	
	private void addAncestorsToMap(Map qt2ReqFields, Field[] anc) {
		if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: registered previous, nb=" + anc.length);
		for (int i = 0; i < anc.length; i++) {
			QueryTree qt = ((QueryTreeField) anc[i]).getQueryTree();
			if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: treating field " + anc[i] + " of node " + qt);
			ArrayList fields = (ArrayList) qt2ReqFields.get(qt);
			if (fields == null) {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: Field, new Qt exists map");
				fields = new ArrayList(1);
				fields.add(anc[i]);
				qt2ReqFields.put(qt, fields);
			}
			else if (!(fields.contains(anc[i]))) {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: Field, Qt exists in map");
				fields.add(anc[i]);
			}
			else {
				if (debug) log.log(BasicLevel.DEBUG, "addFieldInMap: Field already referenced");
			}
		}
	}

	private ArrayList array2arrayList(Object[] array) {
		ArrayList alist = new ArrayList(array.length);
		for (int i = 0; i < array.length; i++) {
			alist.add(array[i]);
		}
		return alist;
	}
}
