/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.PropagFromNestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.Map;

/**
 * This rule removes the query node which are useless.
 * <p> A node is useless when
 * it does not have any filter and all fields are propagated fields which
 * have only one ancestor.
 *
 * @author S. Chassande-Barrioz
 */
public class DropUselessNodeRule extends BasicRule {
	public class UsedFields {
		public boolean isUseLess;
		public Field[] fields;
	}

	public DropUselessNodeRule() {
		super("DropUselessNodeRule");
	}

	public QueryTree rewrite(QueryTree qt, QueryNode _parent)
			throws MedorException {
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		UsedFields uf = isUseless(qt);
		return (uf.isUseLess
			? ((QueryTreeField) uf.fields[0]).getQueryTree()
			: qt);
	}

	protected UsedFields isUseless(QueryTree qt) throws MedorException {

		UsedFields uf = new UsedFields();
		uf.fields = qt.getTupleStructure().getFields();
		uf.isUseLess = false;
		if (debug) log.log(BasicLevel.DEBUG, "Fields: " + printFields(uf.fields));
		if (qt instanceof QueryLeaf) {
			if (debug) log.log(BasicLevel.DEBUG, "useful");
			if (debug) log.log(BasicLevel.DEBUG, "return " + printFields(uf.fields));
			return uf;
		}
		QueryNode qn = (QueryNode) qt;
		QueryTree[] children = qn.getChildren();
		// Look if the current QueryNode is useless
		uf.isUseLess =
			(children.length == 1) &&
			(qn instanceof FilteredQueryTree
			? qn.getQueryFilter() == null
			: true);

		// All fields must be an instanceof of PropagatedField
		int i = 0;
		while (i < uf.fields.length
			&& uf.fields[i] instanceof PropagatedField
			&& !(uf.fields[i] instanceof PropagFromNestedField)
			&& ((PropagatedField) uf.fields[i]).getPreviousFields().length == 1)
			i++;
		uf.isUseLess &= i == uf.fields.length;

		if (uf.isUseLess) {
			if (debug) log.log(BasicLevel.DEBUG, "useless");
			//The current node is useless then it have only one child
			// => Fetch the field which must be used instead of the current
			UsedFields child = isUseless(children[0]);
			if (debug) log.log(BasicLevel.DEBUG, "child: isUseLess:" + child.isUseLess + " ndfield: " + child.fields.length);
			// Reorganize the children field in the same order that the current
			// propagated fields
			Field[] fields = new Field[uf.fields.length];
			for (i = 0; i < uf.fields.length; i++) {
				QueryTreeField qtf = (QueryTreeField)
					((PropagatedField) uf.fields[i]).getPreviousFields()[0];
				int idx =
					qtf.getQueryTree().getTupleStructure().getFieldRank(qtf);
				if (debug) log.log(BasicLevel.DEBUG, "idx=" + idx);
				fields[i] = child.fields[idx - 1];
			}
			uf.fields = fields;
			//The returned uf contains the fields of the child
		}
		else {
			if (debug) log.log(BasicLevel.DEBUG, "useful");
			// Propagate the call to the method.
			Map childrenUseLess = new HashMap();
			for (i = 0; i < children.length; i++) {
				childrenUseLess.put(children[i], isUseless(children[i]));
			}

			// The current node is not useless but the children can be useless.
			// Replace the use of the old fields by the newer if needed

			//in the fields:
			//--------------
			for (i = 0; i < uf.fields.length; i++) {
				if (debug) log.log(BasicLevel.DEBUG, "useful: field " + uf.fields[i].getName());
				if (uf.fields[i] instanceof PropagatedField) {
					//Note: The PropagFromNestedField is also PropagatedField
					PropagatedField pf = (PropagatedField) uf.fields[i];
					Field[] ancs = pf.getPreviousFields();
					QueryTreeField[] neoAncs = new QueryTreeField[ancs.length];
					boolean mustReplace = false;
					for (int j = 0; j < ancs.length; j++) {
						if (debug) log.log(BasicLevel.DEBUG, "useful: field " + uf.fields[i].getName() + " / ancestor: " + ancs[j].getName());
						if (debug) log.log(BasicLevel.DEBUG, "-> for QueryTree " + qt);
						QueryTree qtof =
							((QueryTreeField) ancs[j]).getQueryTree();
						UsedFields childUF =
							(UsedFields) childrenUseLess.get(qtof);
						if (debug) log.log(BasicLevel.DEBUG, "childUF.fields " + childUF.fields.length);
						//if (debug) QueryTreePrinter.printQueryTree(qtof, log);
						if (childUF != null) {
							int idx =
								qtof.getTupleStructure().getFieldRank(ancs[j]);
							mustReplace = true;
							neoAncs[j] = (QueryTreeField) childUF.fields[idx - 1];
						}
						else
							neoAncs[j] = (QueryTreeField) ancs[j];
					}
					if (mustReplace) {
						if (debug) log.log(BasicLevel.DEBUG, "replace " + printFields(pf.getPreviousFields()) + " by " + printFields(neoAncs));
						qn.updatePropagatedField(pf.getName(), neoAncs);
					}
				}
				else if (uf.fields[i] instanceof CalculatedField) {
					Expression e =
						((CalculatedField) uf.fields[i]).getExpression();
					if (replaceInFilter(e, childrenUseLess))
						qn.updateCalculatedField(uf.fields[i].getName(), e);
				}
			}


			//in the filter:
			//--------------
			if (qn instanceof FilteredQueryTree) {
				Expression e = qn.getQueryFilter();
				if (replaceInFilter(e, childrenUseLess))
					qn.setQueryFilter(e);
			}

			//The returned uf contains the fields of the current node
		}

		if (debug) log.log(BasicLevel.DEBUG, "return " + printFields(uf.fields));
		return uf;
	}

	protected boolean replaceInFilter(Expression e, Map map)
		throws MedorException {
		if (e instanceof Operator) {
			boolean result = false;
			for (int i = 0; i < ((Operator) e).getOperandNumber(); i++) {
				result |= replaceInFilter(((Operator) e).getExpression(i), map);
			}
			return result;
		}
		else if (e instanceof FieldOperand) {
			QueryTreeField qtf = (QueryTreeField) ((FieldOperand) e).getField();
			Field neo = replaceField(qtf, map);
			if (neo != qtf) {
				((FieldOperand) e).setField(neo);
				if (debug) log.log(BasicLevel.DEBUG, "Replace in FieldOperand " + qtf.getName() + " by " + neo.getName());
				return true;
			}
		}
		return false;
	}

	protected Field replaceField(QueryTreeField qtf, Map map)
		throws MedorException {
		QueryTree qtChild = qtf.getQueryTree();
		UsedFields childUF = (UsedFields) map.get(qtChild);
		if (childUF != null) {
			int idx = qtChild.getTupleStructure().getFieldRank(qtf);
			return childUF.fields[idx - 1];
		}
		else
			return qtf;
	}

	private String printFields(Field[] fs) {
		String m = "";
		for (int i = 0; i < fs.length; i++) {
			m += (i == 0 ? "" : ",") + fs[i].getName();
		}
		return m;
	}

}
