/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Nest;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class FlattenQueryTreeRule extends BasicRule {

	public FlattenQueryTreeRule() {
		super("FlattenQueryTreeRule");
	}

	public QueryTree rewrite(QueryTree qt, QueryNode parent)
			throws MedorException {
		return flatten(parent, qt);
	}

	public QueryTree flatten(QueryNode parent, QueryTree qt)
			throws MedorException {
		if (qt instanceof QueryLeaf) {
			return qt;
		}
		if (qt instanceof Nest) {
			flatten((Nest) qt, ((QueryNode) qt).getChildren()[0]);
			return qt;
		}
		QueryNode qn = (QueryNode) qt;
		QueryTree[] children = qn.getChildren();
		//Map of replaced fields, for propagation to the parent node
		Map localyReplaced = (parent == null ? null : new HashMap());
		Set inner = findInner(qn);
		while (containsQueryNode(children)) {
			log.log(BasicLevel.DEBUG, "node");
			if (useParent(qn.getQueryFilter())) {
				qn.setQueryFilter(qn.getQueryFilter());
			}
			Field[] fs = qt.getTupleStructure().getFields();
			for (int i = 0; i < fs.length; i++) {
				if (fs[i] instanceof PropagatedField) {
					QueryTreeField childF = (QueryTreeField)
							((PropagatedField) fs[i]).getPreviousFields()[0];
					if (childF.getQueryTree() instanceof QueryLeaf) {
						continue;
					}
					if (childF instanceof PropagatedField) {
						qn.updatePropagatedField(fs[i].getName(),
								new QueryTreeField[]{(QueryTreeField)
								((PropagatedField) childF).getPreviousFields()[0]});
					} else if (childF instanceof CalculatedField) {
						//replace the current field fs[i] by childF
						qn.replace((QueryTreeField) fs[i], childF);
						if (parent != null) {
							localyReplaced.put(fs[i], childF);
						}
					} else {
						throw new MedorException("Unmanaged child field: "
								+ childF);
					}
				} else if (fs[i] instanceof CalculatedField) {
					CalculatedField cf = (CalculatedField) fs[i];
					if (useParent(cf.getExpression())) {
						qn.updateCalculatedField(
								cf.getName(), cf.getExpression());
					}
				} else {
					throw new MedorException("Unmanaged Field: " + fs[i]);
				}
			}
			//Replace the usage of replaced fields in the parent
			replaceUsage(parent, localyReplaced);

			//managing the order by clause
			//TODO case where the orderField is already on a QueryLeaf
			OrderField[] orderFields = qt.getOrderBy();
			if (orderFields != null) {
				for (int i = 0; i < orderFields.length; i++) {
					QueryTreeField orderField = orderFields[i].getField();
					if (debug) {
						log.log(BasicLevel.DEBUG, "managing orderField " + orderField);
					}
					//nothing to do if orderField already on a leaf
					if (orderField.getQueryTree() instanceof QueryLeaf) {
						if (debug) {
							log.log(BasicLevel.DEBUG, "nothing to do for this field (already on a leaf) " + orderField);
						}
						continue;
					}
					if (orderField instanceof PropagatedField) {
						QueryTreeField childF = (QueryTreeField)
								((PropagatedField) orderField)
								.getPreviousFields()[0];
						if (childF.getQueryTree() instanceof QueryLeaf) {
							continue;
						}
						if (childF instanceof PropagatedField) {
							//get down one level
							orderFields[i].setField(
									(QueryTreeField)
									((PropagatedField) childF)
									.getPreviousFields()[0]);
						} else if (childF instanceof CalculatedField) {
							//replace current orderField with childF
							orderFields[i].setField(childF);
						} else {
							throw new MedorException("Unmanaged child field: "
									+ childF);
						}
					} else if (orderField instanceof CalculatedField) {
						CalculatedField cf = (CalculatedField) orderField;
						if (useParent(cf.getExpression())) {
							//replace current orderField expression
							((CalculatedField) orderField)
									.setExpression(cf.getExpression());
						}
					} else {
						throw new MedorException("Unmanaged Field: " + orderField);
					}
				}
				qt.setOrderBy(orderFields);
			}

			//fetch the filter of my children
			boolean modified = false;
			Expression e = qn.getQueryFilter();
			for (int i = 0; i < children.length; i++) {
				if (children[i] instanceof FilteredQueryTree) {
					Expression filter = ((FilteredQueryTree) children[i])
							.getQueryFilter();
					if (filter != null) {
						if (e == null) {
							e = filter;
						} else {
							e = new org.objectweb.medor.expression.lib.And(e, filter);
						}
						modified = true;
					}
				}
			}
			if (modified) {
				log.log(BasicLevel.DEBUG, "Filter modified: " + ExpressionPrinter.e2str(e));
				qn.setQueryFilter(e);
			}
			children = qn.getChildren();
		}
		for (int i = 0; i < children.length; i++) {
			qn.setOuter(children[i], !inner.contains(children[i]));
		}
		return qt;
	}

	private Set findInner(QueryTree qt) {
		Set result = new HashSet();
		findInner(qt, false, result);
		return result;
	}

	private void findInner(QueryTree qt, boolean inner, Set result) {
		if (inner) { result.add(qt); }
		if (qt instanceof QueryNode) {
			QueryNode qn = (QueryNode) qt;
			QueryTree[] children = qn.getChildren();
			for (int i = 0; i < children.length; i++) {
				findInner(children[i], inner || !qn.isOuter(children[i]), result);
			}
		}
        }

	private boolean containsQueryNode(QueryTree[] qts) {
		log.log(BasicLevel.DEBUG, "Entering containsQueryNode for " + qts.length + " QueryTrees.");
		for (int i = 0; i < qts.length; i++) {
			if (qts[i] instanceof QueryNode) {
				log.log(BasicLevel.DEBUG, "Found QueryNode " + qts[i]);
				return true;
			} else {
				log.log(BasicLevel.DEBUG, "Found other QueryTree " + qts[i]);
			}
		}
		log.log(BasicLevel.DEBUG, "containsQueryNode: no QueryNode found. Exiting");
		return false;
	}

	private boolean useParent(Expression e) {
		if (e instanceof Operator) {
			int size = ((Operator) e).getOperandNumber();
			boolean res = false;
			for (int i = 0; i < size; i++) {
				res |= useParent(((Operator) e).getExpression(i));
			}
			return res;
		} else if (e instanceof FieldOperand) {
			FieldOperand fo = (FieldOperand) e;
			QueryTreeField f = (QueryTreeField) fo.getField();
			if (f.getQueryTree() instanceof QueryLeaf) {
				return false;
			}
			if (f instanceof PropagatedField) {
				fo.setField(((PropagatedField) f).getPreviousFields()[0]);
				return true;
			} else {
				log.log(BasicLevel.ERROR, "Unmanaged FieldOperand: " + f);
			}
		}
		return false;
	}

}
