/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.api;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.tuple.api.TupleLoader;

/**
 * @author Sebastien Chassande-Barrioz
 */
public interface QueryNode extends FilteredQueryTree {

	/**
	 * Creates and adds a PropagatedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added. If null, the name is
	 * computed from the origin Fields (parameter anc).
	 * @param type is the PType of the Field to be added. If null, the type is
	 * computed from the origin Fields (parameter anc).
	 * @param anc are the origin Fields for the PropagatedField to be created.
	 * @return the newly created PropagatedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	PropagatedField addPropagatedField(String name, PType type,
									  QueryTreeField[] anc)
		throws MedorException;

	/**
	 * Creates and adds a CalculatedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added.
	 * @param type is the PType of the Field to be added.
	 * @param exp is the Expression for calculating the new Field.
	 * @return the newly created CalculatedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	CalculatedField addCalculatedField(String name, PType type,
									  Expression exp)
		throws MedorException;

	/**
	 * Removes a Field.
	 * @param name is the name of the Field to be removed.
	 * @return the Field which has been removed, or null if no such Field could
	 * be removed.
	 * @throws MedorException when no Field with such name exist.
	 */
	QueryTreeField removeField(String name)
		throws MedorException;

	/**
	 * It retreives the array of the QueryTree used by this QueryNode.
	 */
	QueryTree[] getChildren();

	/**
	 * Indicates whether a row of nulls should be inserted for the
	 * given child if the filter condition fails for all rows in
	 * that child.
	 **/
	boolean isOuter(QueryTree child);

	void setOuter(QueryTree child, boolean outer);

	/**
	 * It retrieves the operation type of the current QueryNode.
	 */
	short getType();

	/**
	 * An integer with position i in this array gives the position of the
	 * corresponding field (field number i of the TupleStructure of this
	 * QueryNode) in the array made of the concatenation of the array of fields
	 * from all children of this QueryNode.
	 */
	//int[] getFieldIndexes();

	/**
	 * Returns the TupleLoader currently associated with this QueryNode.
	 * Note that TupleLoaders are only used at evaluation time.
	 */
	TupleLoader getTupleLoader();

	/**
	 * Associates a TupleLoader with this QueryNode.
	 */
	void setTupleLoader(TupleLoader loader);

    void updatePropagatedField(String name, QueryTreeField[] previous)
        throws MedorException;

    void updateCalculatedField(String name, Expression e) throws MedorException;

    QueryTreeField replace(QueryTreeField old, QueryTreeField neo);
}
