/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */
package org.objectweb.medor.query.lib;

import java.util.List;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The MemberOfHelper class offers methods to manipulate MemberOf operators.
 */
public class MemberOfHelper {

	public static Logger logger = Log.getLoggerFactory()
	.getLogger("org.objectweb.medor.query.lib.MemberOfHelper");

	/**
	 * Adds a MemberOf to a query node, given the left and right ArrayLists.
	 * <p>
	 * If the QueryNode already has an expression, the MemberOf is appended with
	 * and And.
	 * <p>
	 * The field operands may, in this case, be on the same (lower level) query
	 * node. If this is the case, this method tries to un-propagates the right
	 * propagated field if possible, and places the MemberOf on this (lower
	 * level) query node.
	 * <p>
	 * It is assumed that all left fields are on the same query node, and all
	 * right fields are on the same query node.
	 * 
	 * @param qn the QueryNode on which to add the MemberOf
	 * @param mofLeft the ArrayList of field operands for the left part of the
	 *            MemberOf
	 * @param mofRight the ArrayList of field operands for the right part of the
	 *            MemberOf
	 * @throws MalformedExpressionException returned by the MemberOf
	 *             construction
	 */
	public static void addMemberOf(QueryNode qn, List mofLeft, List mofRight)
			throws MedorException, MalformedExpressionException {
		int size = mofLeft.size();
		Expression left, right;
		for (int i = 0; i < size; i++) {
			logger.log(BasicLevel.DEBUG, "Handling field number " +i);
			left = (Expression) mofLeft.get(i);
			right = (Expression) mofRight.get(i);
			if (left instanceof FieldOperand) {
				QueryTreeField leftField = (QueryTreeField) ((FieldOperand) left)
						.getField();
				QueryTreeField rightField = (QueryTreeField) ((FieldOperand) right)
						.getField();
				qn = pushDown(leftField, rightField, mofLeft, mofRight, qn, i);
			}
		}
		MemberOf mof = new MemberOf(mofLeft, mofRight);

		if (qn.getQueryFilter() != null) {
			qn.setQueryFilter(new And(qn.getQueryFilter(), mof));
		} else {
			qn.setQueryFilter(mof);
		}
	}

	/**
	 * Recursive method which goes down the tree, and stops as soon as the two
	 * fields are not on the same query node.
	 * @param leftField
	 * @param rightField
	 * @param mofLeft
	 * @param mofRight
	 * @param qn
	 * @param i the index of the field in the MemberOf lists
	 * @return the query node on which to add the MemberOf operator
	 * @throws MalformedExpressionException
	 * @throws MedorException
	 */
	private static QueryNode pushDown(QueryTreeField leftField,
			QueryTreeField rightField,
			List mofLeft,
			List mofRight,
			QueryNode qn,
			int i) throws MalformedExpressionException, MedorException {
		logger.log(BasicLevel.DEBUG, "Entering pushDown for fields " + 
				leftField.getName() + " and " + rightField.getName());
		QueryTree qt1 = leftField.getQueryTree();
		QueryTree qt2 = rightField.getQueryTree();
		if (!(qt1 instanceof QueryLeaf) && (qt1 == qt2)) {
			logger.log(BasicLevel.DEBUG, "Fields on the same tree " + qt1);
			//both on the same query node: try to go down one level
			if ((rightField instanceof PropagatedField)
					&& (leftField instanceof PropagatedField)) {
				QueryTreeField leftFDown = (QueryTreeField) ((PropagatedField) leftField)
						.getPreviousFields()[0];
				QueryTreeField rightFDown = (QueryTreeField) ((PropagatedField) rightField)
						.getPreviousFields()[0];
				QueryTree qtLDown = leftFDown.getQueryTree();
				if (qtLDown != rightFDown.getQueryTree()) {
					//the two fields are not on the same tree: proceed
					logger.log(BasicLevel.DEBUG, "Propagated from fields on different trees: stop recursion");
					mofLeft.set(i,
							new BasicFieldOperand(((PropagatedField) leftField)
									.getPreviousFields()[0]));
					mofRight.set(i,
							new BasicFieldOperand(
									((PropagatedField) rightField)
											.getPreviousFields()[0]));
					//unpropagate the right field
					((QueryNode) qt2).removeField(rightField.getName());
					//the MemberOf will be attached to the lower node
					return (QueryNode) qt1;
				} else {
					logger.log(BasicLevel.DEBUG, "Propagated from fields on the same tree: go on with recursion");
					//the two fields are still on the same tree: go down
					QueryNode qtPushed = pushDown(leftFDown, rightFDown, mofLeft, mofRight,
							(QueryNode) qtLDown, i);
					//unpropagate the right field
					((QueryNode) qt2).removeField(rightField.getName());
					return qtPushed;
				}
			} else {
				throw new MalformedExpressionException(
						"The helper cannot deal with two fields on the same query node which are not propagated fields");
			}
		} else {
			return qn;
		}
	}

}