/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryNodeException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.OperationType;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Map;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class Nest extends BasicQueryNode implements NestQueryNode {

    /**
     * It lists the fields which have been grouped.
     */
    protected PropagatedField[] myGroupByFields = null;
    protected QueryTreeField[] theGroupByFields = null;
    protected NestedField nestedField = null;

	public Nest() {
	}

	/**
	 * Creates a QueryNode corresponding to a nest operation (group by).
	 * The Fields that are grouped and those that are used for the grouping
	 * must come from one single (child) QueryTree.
	 * @param groupedFields is the array of Fields which are nested by the
	 * operator. These are Fields of the child QueryTree.
	 * @param groupedFieldName is the name of the Field corresponding to the
	 * nesting.
	 * @param groupByFields is the array of Fields which are used for defining
	 * the group to be nested (equivalent to SQL GROUP BY). These are Fields of
	 * the child QueryTree. They are also called NestingFields.
	 */
    public Nest(QueryTreeField[] groupedFields,
                String groupedFieldName,
                QueryTreeField[] groupByFields,
                String nodeName) throws MedorException{
        super(nodeName);
		boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		if (debug) logger.log(BasicLevel.DEBUG, "Nest. Super() done. Entering core.");
		//adding all groupBy fields as PropagatedFields
		
 		this.myGroupByFields = new PropagatedField[groupByFields.length];
 		this.theGroupByFields = new QueryTreeField[groupByFields.length];
		
		for (int i = 0; i < groupByFields.length; i++) {
/*
			QueryTreeField[] anc = {groupByFields[i]};
			if (debug) logger.log(BasicLevel.DEBUG, "Adding GroupByField " + groupByFields[i] + " of node " + groupByFields[i].getQueryTree());
			this.myGroupByFields[i] = this.addPropagatedField(
				groupByFields[i].getName(),
				groupByFields[i].getType(),
				anc);
				*/
			this.theGroupByFields[i] = groupByFields[i];
		}
		
		//creating and adding the NestedField
        this.replaceNestedField(groupedFieldName, groupedFields);
		if (debug) logger.log(BasicLevel.DEBUG, "Checking number of children for " + this);
		QueryTree[] theChildren = this.getChildren();
		if (debug) logger.log(BasicLevel.DEBUG, "Found " + theChildren.length + " chid(ren) for Nest node");
		if (theChildren.length > 1) {
			throw new QueryNodeException("Too many children for this Nest node");
		}
     }

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		Nest n = (Nest) clone;
		n.nestedField = (NestedField) getClone(nestedField, obj2clone);
		if (myGroupByFields != null) {
			n.myGroupByFields = new PropagatedField[myGroupByFields.length];
			for(int i=0; i<myGroupByFields.length; i++) {
				n.myGroupByFields[i] = (PropagatedField) getClone(myGroupByFields[i], obj2clone);
			}
		}
		return clone;
	}

	/**
	 * Creates and adds a NestedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added.
	 * @param grouped are the grouped Fields for the NestedField to be created.
	 * @return the newly created NestedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	public NestedField replaceNestedField(String name,
									  Field[] grouped)
		throws MedorException {
		//removing the old one it it exists
		if (nestedField != null) {
			this.removeField(nestedField.getName());
		}
		NestedField newField = new BasicNestedField(name, this, grouped);
		this.addField(newField);
		nestedField = newField;
		updateChildren();
		return newField;
	}

	public void setQueryFilter(Expression e) {
		throw new UnsupportedOperationException("Nest query nodes cannot be assigned a filter.");
	}

    public QueryTreeField[] getNestingFields() {
        return theGroupByFields;
    }

    public QueryTreeField getNestingField(String fieldName) {
        for (int i = 0; i < theGroupByFields.length; i++) {
            if (theGroupByFields[i].getName().equals(fieldName)) {
                return theGroupByFields[i];
            }
        }
        return null;
    }
    
    public NestedField getNestedField() {
        return nestedField;
    }

    public QueryTreeField replaceNestingField(QueryTreeField old, QueryTreeField neo) {
		boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
    		for (int i = 0; i < theGroupByFields.length; i++) {
    			if (theGroupByFields[i].equals(old)) {
    				theGroupByFields[i] = neo;
    				return neo;
    			}
    		}
    		logger.log(BasicLevel.DEBUG, "Field not found " + old);
    		return null;
    }
    
    public short getType() {
        return OperationType.NEST;
    }
}
