/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author S. Chassande-Barrioz
 */
public class QueryTreePrinter {

	static int id = 0;

	public static void printQueryTree(QueryTree qt, Logger log, int level) {
		if (log != null && log.isLoggable(level)) {
			id = 0;
			printQT("", qt, log, level);
		}
	}

	public static void printQueryTree(QueryTree qt, Logger log) {
		if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
			id = 0;
			printQT("", qt, log, BasicLevel.DEBUG);
		}
	}

	private static void printQT(String prefix, QueryTree qt, Logger log,
								int level) {
		id++;
		int myId = id;
		log.log(level, prefix + "*Node number " + myId);
		log.log(level, prefix + " - Class " + qt);
		log.log(level, prefix + " - Name " + qt.getName());
		if (qt instanceof FilteredQueryTree) {
			Expression e = ((FilteredQueryTree) qt).getQueryFilter();
			if (e != null)
				log.log(level, prefix + " -Filter: " + ExpressionPrinter.e2str(e));
		}
		if (qt instanceof RdbQueryLeaf) {
			try {
				log.log(level, prefix + " - sqlRequest: " +
						((RdbQueryLeaf) qt).getSqlRequest(
								null,
								false,
								false));
			}
			catch (Exception e) {
				log.log(level, prefix + " - sqlRequest: error");
				e.printStackTrace(System.out);
			}
			if (qt instanceof RdbExpQueryLeaf) {
				log.log(level, prefix + "(resetting SQL request)");
				((RdbExpQueryLeaf) qt).resetSqlRequest();
				QualifiedTable[] qts = ((RdbExpQueryLeaf) qt).getQualifiedTables();
				for (int i = 0; i < qts.length; i++) {
					log.log(level, prefix + " - tableName[" + i + "]:"
						+ qts[i].getTableName() + " as " + qts[i].getAliasName());
				}
			}
		}
		log.log(level, prefix + " - Field names:");

		Field[] fs = qt.getTupleStructure().getFields();
		String msg = null;
		for (int i = 0; i < fs.length; i++) {
			msg = prefix + "\t-" + fs[i].getName();
			if (fs[i] instanceof RdbField) {
				msg += "(" + ((RdbField) fs[i]).getColumnName();
				if (fs[i] instanceof RdbExpField)
					msg += " in "
						+ ((RdbExpField) fs[i]).getTable().getTableName()
						+ " as "
						+ ((RdbExpField) fs[i]).getTable().getAliasName();
				msg += ")";
			}
			else if (fs[i] instanceof CalculatedField) {
				msg += " expression=" + ExpressionPrinter.e2str(((CalculatedField) fs[i]).getExpression());
			}
			else if (fs[i] instanceof PropagatedField) {
				Field[] previous = ((PropagatedField) fs[i]).getPreviousFields();
				msg += " previous=";
				for (int j = 0; j < previous.length; j++) {
					msg += previous[j].getName() + " ";
					msg += "of node " + ((QueryTreeField)previous[j]).getQueryTree();
				}
			}
			log.log(level, msg);
		}
		
		if (qt instanceof Nest) {
			log.log(level, prefix + " --Nest: group by fields");
			Field[] gbFields = ((Nest)qt).getNestingFields();
			for (int i = 0; i < gbFields.length; i++) {
			    	msg = " ---groupBy:" + gbFields[i];
				if (gbFields[i] instanceof PropagatedField) {
					Field[] previous = ((PropagatedField) gbFields[i]).getPreviousFields();
					msg += " previous=";
					for (int j = 0; j < previous.length; j++) {
						msg += previous[j].getName() + " ";
						msg += "of node " + ((QueryTreeField)previous[j]).getQueryTree();
					}
				}
				log.log(level, msg);
			}
		}
        OrderField[] ofs = qt.getOrderBy();
        if (ofs != null) {
            StringBuffer sb = new StringBuffer();
            String sep = "";
            for(int i=0; i<ofs.length; i++) {
                sb.append(sep);
                sep = ", ";
                sb.append(ofs[i].getField().getName());
                sb.append((ofs[i].getDesc() ? " descending" : " ascending"));
            }
            log.log(level, prefix + " - Order By: " + sb.toString());
        }
		log.log(level, "");
		if (qt instanceof QueryNode) {
			for (Iterator it = getChildren((QueryNode) qt).iterator(); it.hasNext();) {
				log.log(level, prefix + " -Children of " + myId);
				QueryTree qtchild = (QueryTree) it.next();
				if (qt instanceof QueryNode) {
					log.log(level, prefix + "\t" +
						(((QueryNode) qt).isOuter(qtchild) ? "OUTER" : "INNER"));
				}
				log.log(level, prefix + "\t---------------");
				if (qtchild == null)
					log.log(level, prefix + "\tNull child !!!!!!!");
				else
					printQT(prefix + "\t", qtchild, log, level);
			}
		}
	}

	/**
	 * Updates the list of children by screening all Fields and the
	 * query filter.
	 */
	public static ArrayList getChildren(QueryNode qn) {
		ArrayList children = new ArrayList();
		Field[] fs = qn.getTupleStructure().getFields();
		//add children from Fields
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				Field[] anc = ((PropagatedField) fs[i]).getPreviousFields();
				for (int j = 0; j < anc.length; j++) {
					QueryTree qt = ((QueryTreeField) anc[j]).getQueryTree();
					//System.log.log(level, "PropagatedField " + fs[i].getName() + ": anc["+j+"]=" + anc[j].getName() + " , qt=" + qt);
					if (!children.contains(qt)) {
						children.add(qt);
					}
				}
			}
			else if (fs[i] instanceof NestedField) {
				Field[] grouped = ((NestedField) fs[i]).getFields();
				for (int j = 0; j < grouped.length; j++) {
					QueryTree qt = ((QueryTreeField) grouped[j]).getQueryTree();
					//System.log.log(level, "NestedField " + fs[i].getName() + ": grouped["+j+"]=" + grouped[j].getName() + " , qt=" + qt);
					if (!children.contains(qt)) {
						children.add(qt);
					}
				}
			}
			else if (fs[i] instanceof CalculatedField) {
				addChildrenFromExpression(children, ((CalculatedField) fs[i]).getExpression());
			}
		}
		addChildrenFromExpression(children, qn.getQueryFilter());
		return children;
	}

	/**
	 * Parses an Expression and adds the QueryTrees of the FieldOperands to
	 * the children of the current QueryNode.
	 * @param exp is the Expression to be parsed.
	 */
	private static void addChildrenFromExpression(ArrayList children, Expression exp) {
		if (exp instanceof FieldOperand) {
			Field f = ((FieldOperand) exp).getField();
			QueryTree qt = ((QueryTreeField) f).getQueryTree();
			//System.log.log(level, "fieldOperand " + f.getName() + " , qt=" + qt);
			if (!children.contains(qt)) {
				children.add(qt);
			}
		}
		else if (exp instanceof Operator) {
			for (int i = 0; i < ((Operator) exp).getOperandNumber(); i++) {
				addChildrenFromExpression(children, ((Operator) exp).getExpression(i));
			}
		}
	}

}
