/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.rdb.api;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import java.util.List;

/**
 * An RdbExpQueryLeaf is a particular type of RdbQueryLeaf for which the
 * associated SQL query if not given as a SQL String, but is computed from
 * an Expression (representing the filter) and from abstract information
 * made of QualifiedTables, representing the FROM part of the SQL query.
 * <p>
 * The Medor optimizer can "reason" about RdbExpQueryLeaves, and merge several
 * such query leaves into one if they have the same DataStore.
 * <p>
 * The life cycle of an RdbExpQueryLeaf is:
 * <ul><li>create the leaf (constructor)</li>
 * <li>add the fields (addRdbField)</li>
 * <li>add the expression which is built on the fields</li>
 * </ul>
 *
 * A RdbExpQueryLeaf can use a RdbAdapter to build a SQL query. The Rdb Adapter
 * permits to manage the relational databases that are not conform to the jdbc
 * specification.
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 *			Alexandre Lefebvre
 * 			</b></A>
 */
public interface RdbExpQueryLeaf extends RdbQueryLeaf, FilteredQueryTree {
    /**
     * Adds a Field to the RdbExpQueryLeaf
     * @param fieldName is the name of the Field to be added.
     * @param type is the PTyme of the Field to be added.
     * @param colName is the name of the column in the associated
     * QualifiedTable.
     * @param table is the QualifiedTable from which the Field is to be created.
     * @return the created RdbExpField.
     * @throws QueryLeafException if the QualifiedTable is not in the
     * set of QualifiedTables associatged with the current RdbExpQueryLeaf.
     */
    RdbExpField addRdbField(String fieldName,
                            PType type,
                            String colName,
                            QualifiedTable table) throws QueryLeafException;

	/**
     * Adds an RdbExpField to the RdbExpQueryLeaf. The QueryLeaf associated
	 * to the RdbExpField becomes the current RdbExpQueryLeaf.
     * @param rdbField is the RdbField to be added.
     */
    RdbExpField addRdbField(RdbExpField rdbField)
		throws QueryLeafException;

	/**
     * Adds an RdbExpField to the RdbExpQueryLeaf. The QueryLeaf associated
	 * to the RdbExpField becomes the current RdbExpQueryLeaf.
     * @param rdbField is the RdbField to be added.
     */
    void removeRdbField(RdbExpField rdbField)
		throws QueryLeafException;

    /**
     * Adds a groupBy Field. This RdbExpField must be in The TupleStructure of
     * this QueryLeaf Object.
     * @param groupBy : RdbExpField
     * @throws QueryLeafException if this RdbExpField doesn't appears in
     * the list of the fields which will be projected in Select clause
     */
    void addGroupBy(RdbExpField groupBy) throws QueryLeafException;

     /**
     * Gets the RdbExpFields of the GroupBy Clause.
     * It returns null if there is no Result Grouping
     * @return an array of RdbExpField
     */
    RdbExpField[] getGroupByFields();

    /**
     * Returns the associated QualifiedTables.
     * @return the array of associated QualifiedTables.
     */
    QualifiedTable[] getQualifiedTables();

	/**
	 * Adds a QualifiedTable to the QueryLeaf
	 * @param qt the QualifiedTable to be added
	 */
	void addQualifiedTable(QualifiedTable qt);

	/**
	 * Adds a QualifiedTable to the QueryLeaf
	 * @param qt the QualifiedTable to be added
	 * @throws MedorException if the QualifiedTable in not in the current leaf.
	 */
	void removeQualifiedTable(QualifiedTable qt)
	throws MedorException;

	/**
	 * Checks if a QualifiedTable is in the list of QualifiedTables of this
	 * QueryLeaf
	 * @param qt the QualifiedTable to be tested
	 * @return true if qt is in the list of QualifiedTables for this QueryLeaf
	 */
	boolean containsQualifiedTable(QualifiedTable qt);

	/**
	 * It assignes the rdb adapter name.
	 * The RdbAdapter is used to adapte the sql queries to a relational database
	 */
	void setRdbAdapterName(String name);

	/**
	 * It retrieves the rdb adapter name.
	 * The RdbAdapter is used to adapte the sql queries to a relational database
	 */
	String getRdbAdapterName();

	/**
	 * It assigns the rdb adapter.
	 * The RdbAdapter is used to adapt sql queries to a relational database
	 */
	void setRdbAdapter(RdbAdapter adapter);

	/**
	 * It retrieves the rdb adapter.
	 * The RdbAdapter is used to adapt sql queries to a relational database
	 */
	RdbAdapter getRdbAdapter();

    void setRootJoinedTables(List rootjts);

	/**
	 * In case the SQL query was already computed, resets it to null.
	 */
	void resetSqlRequest();
	
	/**
	 * Requires that the GROUP BY clause not be output by the getSqlRequest method.
	 * @param noGroupBy is true if no GROUP BY clause should be output (false by
	 * default).
	 */
	void setNoGroupBy(boolean noGroupBy);
	
	/**
	 * Returns the GROUP BY clause of the query leaf.
	 * @return the GROUP BY clause of the query leaf.
	 */
	String getGroupBy();	
	
	/**
	 * Returns the GROUP BY clause of the query leaf.
	 * @return the GROUP BY clause of the query leaf.
	 */
	String getGroupBy(String singleAlias);
}
