/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.query.rdb.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Nest;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;

import java.util.ArrayList;
import java.util.Map;

/**
 * This class represents a QueryLeaf which contains aggregate functions.
 * <p> It is built on top of a RbdQueryLeaf.
 */
public class AggregateRdbQueryNode extends Nest
		implements RdbQueryLeaf, NestQueryNode {

    private static final long serialVersionUID = 1185797036981033368L;

    protected DataStore ds;
	protected RdbExpField groupByField;
	protected RdbExpQueryLeaf subRdbQL;
	protected String query = null;
	private boolean isSubquery = false;

	public AggregateRdbQueryNode() {
	}

	/**
	 * Constructs a RDB QueryNode for an aggregation.
	 * <p>The SQL request (string) will be calculated.
	 * @param ds is the DataStore associated to the QueryLeaf to be created
	 * @throws QueryLeafException if the QualifiedTables have all the same
	 * names, and in this case also the same aliases.
	 */
	public AggregateRdbQueryNode(QueryTreeField[] groupedFields,
                String groupedFieldName,
                QueryTreeField[] groupByFields,
				BasicRdbExpQueryLeaf subQL,
                DataStore ds,
				String nodeName)
			throws MedorException {
		super(groupedFields, groupedFieldName, groupByFields, nodeName);
		this.ds = ds;
		this.subRdbQL = subQL;
		//removing the nested field
		removeField(groupedFieldName);
		//adds the groupby fields
		for (int i = 0; i < groupByFields.length; i++) {
			subQL.addGroupBy((RdbExpField) groupByFields[i]);
		}
	}

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		AggregateRdbQueryNode a = (AggregateRdbQueryNode) clone;
		a.ds = ds;
		a.groupByField = (RdbExpField) getClone(groupByField, obj2clone);
		a.subRdbQL = (RdbExpQueryLeaf) getClone(subRdbQL, obj2clone);
		a.query = query;
		a.isSubquery = isSubquery;
		return clone;
	}

	public String getSqlRequest(ParameterOperand[] pos,
								ArrayList al,
								boolean rangeStartAt,
								boolean rangeSize)
			throws MedorException, ExpressionException {
		if (al == null) {
			return getSqlRequest(pos, rangeStartAt, rangeSize);
		} else {
			throw new MedorException("Request with selectList not supported");
		}
	}

	/**
	 * Builds and returns the SQL query as a String corresponding to the
	 * current RdbExpQueryLeaf.
	 * <p>
	 * The SQL query is computed using the aggregate fields and relies on
	 * the child RdbQueryLeaf.
	 * @return the SQL query as a String.
	 */
	public String getSqlRequest(ParameterOperand[] pos,
	        boolean rangeStartAt,
	        boolean rangeSize)
			throws MedorException, ExpressionException {
		ArrayList selectFields = new ArrayList();
		for (int i = 0; i < fields.size(); i++) {
			if (!(fields.get(i) instanceof NestedField)) {
				if (fields.get(i) instanceof PropagatedField) {
					selectFields.add(((PropagatedField)fields.get(i)).getPreviousFields()[0]);
				} else {
					selectFields.add(fields.get(i));
				}
			}
		}
		if (query != null) {
			return query;
		}
		if (subRdbQL.isSubquery()) {
			//Constructs a subquery only if there is a count on multiple fields.
			//Note: in this case, no other aggregates are supported.
			boolean subRdbQLDistinct = subRdbQL.getDistinct();
			subRdbQL.setDistinct(this.getDistinct());
			StringBuffer sb = new StringBuffer();
			sb.append(subRdbQL.getSelectList("SELECT ", selectFields, false));
			subRdbQL.setDistinct(subRdbQLDistinct);
			subRdbQL.setNoGroupBy(true);

			sb.append(" FROM (");
			sb.append(subRdbQL.getSqlRequest(pos, null, rangeStartAt, rangeSize));
			sb.append(")");
			if (subRdbQL.getRdbAdapter() != null) {
			    String subQueryAlias = subRdbQL.getRdbAdapter().getSubqueryAlias();
			    if (subQueryAlias == null)
			        sb.append(subRdbQL.getGroupBy());
			    else {
			        subRdbQL.getRdbAdapter().writeTableAlias(subQueryAlias, sb);
			        sb.append(subRdbQL.getGroupBy(subQueryAlias));
			    }
				  
			}
			query = sb.toString();
		} else {
			query = subRdbQL.getSqlRequest(pos,
					selectFields,
					rangeStartAt,
					rangeSize);
		}
		return query;
	}

	public String getSqlDelete(ParameterOperand[] pos) {
        // TODO Auto-generated method stub
        return null;
    }

    public String getSqlUpdate(ParameterOperand[] pos, Expression updateExpression) {
        // TODO Auto-generated method stub
        return null;
    }

    public String getSelectList(String selectList, ArrayList selectFields,
								boolean qualified)
		throws MedorException {
		throw new QueryLeafException("Unsupported method getSelectList for AggregateRdbQueryNode");
	}

	public DataStore getDataStore() {
		return ds;
	}

	public void setIsSubquery(boolean subquery) {
		isSubquery = subquery;
	}

	public boolean isSubquery() {
		return isSubquery;
	}

	/**
	 * Redefines getQueryFilter using the filter of the subquery.
	 * @return the filter of the subquery.
	 */
	public Expression getQueryFilter() {
		if (subRdbQL != null) {
			return subRdbQL.getQueryFilter();
		} else {
			return null;
		}
	}
}