/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.query.rdb.lib;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.CalculatedParameterOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.UnaryOperator;
import org.objectweb.medor.expression.converter.rdb.Expression2WhereClauseImpl;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.InCollection;
import org.objectweb.medor.filter.lib.IsEmpty;
import org.objectweb.medor.filter.lib.IsNull;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.type.lib.QType;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This class overrides Expression2WhereClauseImpl for MEDOR-specific operators.
 * <p>It is meant to be used for a particular QueryLeaf.<p>
 *
 * @author A. Lefebvre
 */
public class MedorExpression2WhereClause extends Expression2WhereClauseImpl {

    public static final Integer POS = new Integer(1);
    public static final Integer QUERY_LEAF = new Integer(2);
    
    public MedorExpression2WhereClause() {
        super();
    }

	public static String e2where(Expression exp,
										 RdbAdapter rdbAdapter,
										 BasicRdbExpQueryLeaf ql,
										 ParameterOperand[] pos)
	throws ExpressionException {
        Map ctx = new HashMap();
        ctx.put(POS, pos);
        ctx.put(QUERY_LEAF, ql);
		return Expression2WhereClauseImpl.e2where(exp, rdbAdapter, ctx);
	}

    private static ParameterOperand[] getPos(Map ctx) {
        return (ParameterOperand[]) ctx.get(POS);
    }
    private static ParameterOperand getPO(ParameterOperand[] pos, String paramName) {
        if (pos != null) {
            for (int i = 0; i < pos.length; i++) {
                if (paramName.equals(pos[i].getName())) {
                    return pos[i];
                }
            }
        }
        return null;
    }
    
    private BasicRdbExpQueryLeaf getQL(Map ctx) {
        return (BasicRdbExpQueryLeaf) ctx.get(QUERY_LEAF);
    }
    
	public void convertExp2WhereClause(Expression exp,
									   RdbAdapter rdbAdapter,
									   StringBuffer sb,
                                       Map ctx)
			throws ExpressionException {
		if (exp instanceof FieldOperand) {
			sb.append(BasicRdbExpQueryLeaf.getQualifiedFieldName(
					((RdbExpField) ((FieldOperand) exp).getField())));
		} else if (exp instanceof InCollection) {
			InCollection bo = (InCollection) exp;
			sb.append("(");
			convertExp2WhereClause(bo.getExpression(0), rdbAdapter, sb, ctx);
			sb.append(" ");
			sb.append(toSqlOperator(bo.getOperatorString()));
			sb.append(" ");
			getSQLEnumeration(bo.getExpression(1),
							bo.getElemType(),
                            getPos(ctx), sb);
			sb.append(")");
		} else if (exp instanceof MemberOf) {
			MemberOf mo = (MemberOf) exp;
			int size = mo.getOperandNumber() / 2;
			sb.append("(");
			for (int i = 0; i < size; i++) {
				convertExp2WhereClause(mo.getExpression(i), rdbAdapter, sb, ctx);
				sb.append(i < (size - 1) ? ", " : "");
			}
			sb.append(" IN (");
			FieldOperand fop = (FieldOperand) mo.getExpression(size);
			appendSubquery(fop, rdbAdapter, sb, ctx);
			sb.append("))");
		} else if (exp instanceof IsEmpty) {
			IsEmpty ie = (IsEmpty) exp;
			sb.append(" NOT EXISTS (");
			Expression e = ie.getExpression(0);
			if (e instanceof FieldOperand) {
				appendSubquery((FieldOperand) e, rdbAdapter, sb, ctx);
			} else {
				throw new IllegalStateException
					("can't handle operand type: " + e);
			}
			sb.append(")");
                }
		//TODO: in the case of NOT xyz, it may sometimes be better to
		//produce something else, e.g. NOT (toto IS NULL) -> toto IS NOT NULL
		else if (exp instanceof IsNull) { //specific MEDOR
			convertExp2WhereClause(((UnaryOperator) exp).getExpression(0), rdbAdapter, sb, ctx);
			sb.append(" ");
			sb.append(toSqlOperator(((Operator) exp).getOperatorString()));
        } else if (exp instanceof Equal || exp instanceof NotEqual) {
            Operator op = (Operator) exp;
            boolean nullValue_0 = 
                isNullParameterOperand(op.getExpression(0), ctx);
            boolean nullValue_1 = 
                isNullParameterOperand(op.getExpression(1), ctx);
            if (nullValue_0 ^ nullValue_1) {
                //Transform the equality (or not equality) to a is null operator
                Expression e; 
                if (nullValue_0) {
                    e = op.getExpression(1);
                } else {
                    e = op.getExpression(0);
                }
                IsNull isnull = new IsNull(e); 
                if (exp instanceof NotEqual) {
                    isnull.setNot(true);
                }
                convertExp2WhereClause(isnull, rdbAdapter, sb, ctx);
            } else {
                super.convertExp2WhereClause(exp, rdbAdapter, sb, ctx);
            }
        } else
			super.convertExp2WhereClause(exp, rdbAdapter, sb, ctx);
	}

	private void appendSubquery(FieldOperand fop, RdbAdapter rdbAdapter,
				    StringBuffer sb, Map ctx)
		throws ExpressionException {
		QueryTreeField qtf = (QueryTreeField) fop.getField();
		BasicRdbExpQueryLeaf leaf = (BasicRdbExpQueryLeaf) qtf.getQueryTree();
		if (leaf == getQL(ctx)) {
			throw new ExpressionException
				("can't append self as a subquery " + qtf.getName());
		}
		leaf.setRdbAdapter(rdbAdapter);
		try {
			sb.append(leaf.getSqlRequest(null, false, false));
		} catch (MedorException e) {
			throw new ExpressionException(e);
		}
	}

	private void getSQLEnumeration(
			Expression e,
			PType elemType,
			ParameterOperand[] pos,
			StringBuffer sb)
			throws ExpressionException {
		Collection c = null;
		sb.append("(");
		if (e instanceof ParameterOperand) {
			String pn = ((ParameterOperand) e).getName();
			if (pos == null || pos.length == 0)
				throw new ExpressionException("No Parameter");
			int i = 0;
			while (i < pos.length && !pn.equals(pos[i].getName()))
				i++;
			if (i < pos.length)
				c = (Collection) pos[i].getObject();
			else
				throw new ExpressionException("Parameter " + pn + " not found");
		} else
			c = (Collection) ((Operand) e).getObject();
		Iterator it = c.iterator();
		if (!it.hasNext()) {
			throw new ExpressionException("Empty collection for InCollection operator");
		}
		while (it.hasNext()) {
			Object nextit = it.next();
			//case of ParameterOperand
			if (nextit instanceof ParameterOperand) {
				sb.append("?");
			} else if (elemType.equals(PTypeSpace.STRING)) {
				sb.append("'");
				sb.append(nextit);
				sb.append("'");
			} else {
				sb.append(nextit);
			}
			if (it.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append(")");
	}

	public void convertExp2WhereClauseBuilder(Expression exp,
											  String rdbAdapterVarName,
											  StringBuffer sb,
                                              Map ctx)
		throws ExpressionException {
		if (exp instanceof FieldOperand) {
			sb.append(BasicRdbExpQueryLeaf.getQualifiedFieldName(
					((RdbExpField) ((FieldOperand) exp).getField())));
		} else if (exp instanceof InCollection) {
			InCollection bo = (InCollection) exp;
			sb.append("(");
			convertExp2WhereClauseBuilder(bo.getExpression(0), rdbAdapterVarName, sb, ctx);
			sb.append(" ");
			sb.append(toSqlOperator(bo.getOperatorString()));
			sb.append(" ");
			getSQLEnumeration(bo.getExpression(1),
							bo.getElemType(),
							getPos(ctx), sb);
			sb.append(")");
		} else if (exp instanceof MemberOf) {
			throw new ExpressionException("MemberOf Operator not yet supported at generation time");
		} else if (exp instanceof IsNull) { //specific MEDOR
			convertExp2WhereClauseBuilder(((Operator) exp).getExpression(0), rdbAdapterVarName, sb, ctx);
			sb.append(" ");
			sb.append(toSqlOperator(((Operator) exp).getOperatorString()));
		} else
			super.convertExp2WhereClauseBuilder(exp, rdbAdapterVarName, sb, ctx);
	}
    
    public static boolean isNullParameterOperand(Expression e, 
            Map ctx) throws ExpressionException {
        if (e instanceof ParameterOperand) {
            ParameterOperand[] pos = getPos(ctx);
            if (pos == null) {
                return false;
            }
            ParameterOperand po = getPO(pos, ((ParameterOperand) e).getName());
            if (po != null && isJavaObject(e.getType())) {
                if (e instanceof CalculatedParameterOperand) {
                    ((CalculatedParameterOperand) e).evaluate(pos);
                    return ((CalculatedParameterOperand) e).getObject() == null;
                } else {
                    return po.getObject() == null;
                }
            }
        }
        return false;
    }
    
    public static boolean isJavaObject(PType t) {
        switch (t.getTypeCode()) {
        case PType.TYPECODE_BIGDECIMAL:
        case PType.TYPECODE_BIGINTEGER:
        case PType.TYPECODE_DATE:
        case PType.TYPECODE_OBJBYTE:
        case PType.TYPECODE_OBJCHAR:
        case PType.TYPECODE_OBJDOUBLE:
        case PType.TYPECODE_OBJINT:
        case PType.TYPECODE_OBJLONG:
        case PType.TYPECODE_OBJSHORT:
        case PType.TYPECODE_STRING:
        case QType.TYPECODE_COLLECTION:
            return true;
        default:
            return false;
        }
    }

}
