/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.lib;


import org.objectweb.medor.clone.api.Cloneable;
import org.objectweb.medor.clone.lib.BasicCloneable;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;

import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * This interface is an implementation of the TupleCollection interface using
 * Java collections. It can be used when we have obejcts fetched in the memory.
 */

public class ExplicitTupleCollection
	extends BasicCloneable
	implements TupleCollection {

	private TupleStructure schema;
	private List tuples;
	private int cursor;
	boolean top = true;
	boolean end = false;
	boolean closed = false;

	public ExplicitTupleCollection() {
	}

	public ExplicitTupleCollection(TupleStructure schema) throws MedorException {
        this.schema = schema;
        tuples = new LinkedList();
        cursor = 0;
    }

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		ExplicitTupleCollection tc = (ExplicitTupleCollection) clone;
		tc.closed = closed;
		tc.end = end;
		tc.top = top;
		tc.cursor = cursor;
		tc.schema = (TupleStructure) getClone(schema, obj2clone);
		if (tuples != null) {
			tuples = new LinkedList();
			for(Iterator it = tuples.iterator(); it.hasNext();) {
				tc.tuples.add(getClone((Cloneable) it.next(), obj2clone));
			}
		}
		return clone;
	}

	public void close() throws MedorException {
		closed = true;
	}

    public synchronized TupleStructure getMetaData() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return schema;
    }

    public synchronized boolean isEmpty() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        return (tuples.size() == 0);
    }


    public synchronized boolean isLast() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        // Testing if the TC is empty and if we are out of of bound of the TC(cursor == -1)
        if ((cursor == -1) || (tuples.size() == 0))
            throw new MedorException(" No elements fetched in this tupleCollection");
        else
            return (tuples.size() == getRow());
    }


    public synchronized int size() {
        return tuples.size();
    }

    public synchronized boolean next() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (!(tuples.size() == 0)) { // Not empty
            if (cursor == tuples.size()) { // We are at the end of this TC object
                end = true; // the cursor is out of the bound of thIS tC object
                return false;
            } else if (top) {
                cursor = 1; // The first tuple
                top = false;
                return true;
            } else {
                cursor++;
                return true;
            }
        } else
            return false;
    }

    public synchronized boolean previous() throws MedorException {
        boolean prev = true;
        if (!isEmpty()) {
            if (end) {
                end = false;
                prev = true;
            } else if (getRow() == 1) {
                top = true;
                prev = false;
            } else {
                cursor--;
                prev = true;
            }
        } else {
            prev = false;
        }
        return prev;
    }

    public synchronized boolean row(int row) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if ((row >= 1) && (row <= tuples.size())) {
            cursor = row;
            return true;
        } else
            return false;
    }

    public synchronized void first() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (!isEmpty()) {
            end = false;
            top = false;
            cursor = 1;
        }
    }

    public synchronized int getRow() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (isEmpty() || top || end)
            return -1;
        else
            return cursor;
    }

    /**
     * Insert the tuple T at the designated row number of the TupleCollection
     * @param i the first tuple is 1, the second is 2, ...
     * @param T is the tuple to be inserted
     * @throws MedorException if data source access error or invalid tuple number
     */
    public synchronized void insertTuple(int i, Tuple T) throws MedorException {
        if ((i >= 1) && (i <= size()))
            tuples.add((i - 1), T);
        else
            throw new MedorException("its impossible to insert in this cursor: " + i);
    }

    /**
     * Insert the tuple T at the last of this TupleCollection object
     * @param T is the tuple to be inserted
     * @throws MedorException if data source access error
     */
    public synchronized void insertTuple(Tuple T) throws MedorException {
        tuples.add(T);
    }

    /**
     * Remove the tuple of the designated row number from the TupleCollection object
     * @param row the first tuple is 1, the second is 2, ...
     * @throws MedorException if data source access error or invalid tuple number
     */
    public synchronized void removeTuple(int row) throws MedorException {
        if ((row >= 1) && (row <= size())) {
            tuples.remove(row - 1);
            if (tuples.size() == 0) {
                end = false;
                cursor = -1;
            } else
                cursor = row - 1;
        } else
            throw new MedorException("its impossible to remove from this cursor");
    }

    /**
     * Returns the value of the designated row number of the TupleCollection
     * as a Tuple Object.
     * @param i the first tuple is 1, the second is 2, ...
     * @return a Tuple Object
     * @throws MedorException if data source access error
     */
    public synchronized Tuple getTuple(int i) throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if (i > 0 && i <= tuples.size()) {
            return (Tuple) tuples.get(i - 1);
        } else
            throw new MedorException("its impossible de get tuple from this cursor " + i);

    }

    public synchronized Tuple getTuple() throws MedorException {
		if (closed) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
        if ((!isEmpty()) && (!top) && (!end)) {
            return (Tuple) tuples.get(cursor - 1);

        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a Boolean of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a boolean value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized boolean getBoolean(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getBoolean(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a Byte of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a byte value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized byte getByte(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getByte(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.util.Date of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a Date value as SQL type
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized Date getDate(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getDate(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a double of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a double value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized double getDouble(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getInt(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a float of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a float value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized float getFloat(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getFloat(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as an int of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return an integer value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized int getInt(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getInt(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a char of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a character value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized char getChar(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getChar(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a long of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a long value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized long getLong(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getLong(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a short of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a short value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized short getShort(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getShort(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Returns the value of the designated column in the current row of this Tuple
     * as a java.lang.String of java programming language
     * @param i the first column is 1, the second is 2, ...
     * @return a String value
     * @throws MedorException if data source access error or invalid cursor
     */
    public synchronized String getString(int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            return this.getTuple().getString(i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setBoolean(boolean x, int i) throws MedorException {
        //i represent the current row
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setBoolean(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }


    public synchronized void setByte(byte x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setByte(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setDate(Date x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setDate(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setDouble(double x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setDouble(x, i);
            tuples.add(i, upletI);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setFloat(float x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setFloat(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }


    public synchronized void setInt(int x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setInt(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setShort(short x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setShort(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    public synchronized void setString(String x, int i) throws MedorException {
        if (!isEmpty() && !top && !end) {
            Tuple t = this.getTuple();
            MemoryTuple upletI = (MemoryTuple) t;
            upletI.setString(x, i);
        } else
            throw new MedorException("No elements fetched in this tupleCollection");
    }

    /**
     * Some operations of researchs and iterations.
     */

    public synchronized boolean contains(Tuple T) {
        return tuples.contains(T);
    }

    public synchronized void display() throws MedorException {
        TupleCollectionView frame = new TupleCollectionView(this);
        frame.pack();
        frame.setVisible(true);
    }

    public synchronized Iterator iteratorOf(int i) throws MedorException {
        Object o = getTuple().getObject(i);
        if (o instanceof Collection) {
            Collection c = (Collection) o;
            return c.iterator();
        } else {
            throw new MedorException("the Attribute number " + i + " is not a collection Object");
        }
    }

    public synchronized Iterator iteratorAll() throws MedorException {
        return tuples.iterator();
    }
}