/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleLoader;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.lib.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

public class GeneralTupleLoader implements TupleLoader {

	private Logger logger;

	public GeneralTupleLoader(int[] ranges, TupleStructure ts) {
		indexes = ranges;
		this.ts = ts;
		logger = Log.getLoggerFactory().getLogger(getClass().getName());
	}

	//indexes is the array of attribute ranks to be loaded from the source Tuple
	private int[] indexes;
	TupleStructure ts;

	public void loadTuple(Tuple sourceBuffer,
	                      VariableOperand[] temp,
	                      ParameterOperand[] parameters)
	        throws MedorException, ExpressionException  {
		boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		PType type;
		for (int i = 0; (i < temp.length); i++) {
			type = ts.getField(i + 1).getType();
			Expression cField = null;
			if (indexes[i] == -1) {
				//calculated field
				cField = ((CalculatedField) ts.getField(i + 1)).getExpression();
			}
			switch (type.getTypeCode()) {
			case QTypeTuple.TYPECODE_INT:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getInt(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getInt());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i + "]= " + temp[i].getInt());
				}
				break;
			case QTypeTuple.TYPECODE_SHORT:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getShort(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getShort());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i + "]= " + temp[i].getShort());
				}
				break;
			case QTypeTuple.TYPECODE_BYTE:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getByte(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getByte());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getByte());
				}
				break;
			case QTypeTuple.TYPECODE_LONG:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getLong(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getLong());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getLong());
				}
				break;
			case QTypeTuple.TYPECODE_DOUBLE:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getDouble(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getDouble());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getDouble());
				}
				break;
			case QTypeTuple.TYPECODE_BOOLEAN:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getBoolean(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getBoolean());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getBoolean());
				}
				break;
			case QTypeTuple.TYPECODE_FLOAT:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getFloat(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getFloat());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				break;
			case QTypeTuple.TYPECODE_STRING:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getString(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getString());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getString());
				}
				break;
			case QTypeTuple.TYPECODE_CHAR:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getChar(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getChar());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getChar());
				}
				break;
			case QTypeTuple.TYPECODE_DATE:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getDate(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getDate());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getDate());
				}
				break;
				// TO DO: Add TupleCollection type and Tuple Types
			case QTypeTuple.TYPECODE_OBJBOOLEAN:
			case QTypeTuple.TYPECODE_OBJBYTE:
			case QTypeTuple.TYPECODE_OBJCHAR:
			case QTypeTuple.TYPECODE_OBJDOUBLE:
			case QTypeTuple.TYPECODE_OBJFLOAT:
			case QTypeTuple.TYPECODE_OBJINT:
			case QTypeTuple.TYPECODE_OBJLONG:
			case QTypeTuple.TYPECODE_OBJSHORT:
			case QTypeTuple.TYPECODE_BIGDECIMAL:
			case QTypeTuple.TYPECODE_BIGINTEGER:
			default:
				if (indexes[i] != -1) {
					//propagated field
					temp[i].setValue(sourceBuffer.getObject(indexes[i]));
					temp[i].setIsDefined(sourceBuffer.isDefined(indexes[i]));
				} else {
					synchronized(cField) {
						cField.compileExpression();
						Operand op = cField.evaluate(parameters, sourceBuffer);
						temp[i].setValue(op.getObject());
						temp[i].setIsDefined(op.isDefined());
					}
				}
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Value assigned [" + i +"]= " + temp[i].getObject());
				}
				break;
			}
		}
	}

	public int[] getFieldIndexes() {
		return indexes;
	}

	// NO INTEGRITY VERIFING
	public void setFieldIndexes(int[] indexes) {
		this.indexes = indexes;
	}

	public TupleStructure getTupleStructure() {
		return ts;
	}
}
