/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.clone.lib.BasicCloneable;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;

import java.util.Date;
import java.util.Map;
import java.math.BigDecimal;
import java.math.BigInteger;


/**
 * This class represent a tuple interface in a memory. Java array is used to stores fienlds Values.
 * Rather than getXXX() methods, setXXX() allow updating the content of the Tuple.
 */
public class MemoryTuple extends BasicCloneable implements Tuple {

	private VariableOperand[] uplet;

	public MemoryTuple() {
	}

	public MemoryTuple(PType types[]) {
		uplet = new BasicVariableOperand[uplet.length];
		for (int cpt = 0; cpt < types.length; cpt++) {
			uplet[cpt] = new BasicVariableOperand(types[cpt]);
		}
	}

	public MemoryTuple(VariableOperand fields[]) {
		uplet = fields;
	}

	public MemoryTuple(Operand fields[]) throws TypingException {
		PType type;
		uplet = new BasicVariableOperand[fields.length];
		for (int i = 0; i < uplet.length; i++) {
			type = fields[i].getType();
			uplet[i] = new BasicVariableOperand(type);
			switch (type.getTypeCode()) {
			case QTypeTuple.TYPECODE_INT:
				uplet[i].setValue(fields[i].getInt());
				break;
			case QTypeTuple.TYPECODE_SHORT:
				uplet[i].setValue(fields[i].getShort());
				break;
			case QTypeTuple.TYPECODE_BYTE:
				uplet[i].setValue(fields[i].getByte());
				break;
			case QTypeTuple.TYPECODE_LONG:
				uplet[i].setValue(fields[i].getLong());
				break;
			case QTypeTuple.TYPECODE_DOUBLE:
				uplet[i].setValue(fields[i].getDouble());
				break;
			case QTypeTuple.TYPECODE_BOOLEAN:
				uplet[i].setValue(fields[i].getBoolean());
				break;
			case QTypeTuple.TYPECODE_FLOAT:
				uplet[i].setValue(fields[i].getFloat());
				break;
			case QTypeTuple.TYPECODE_STRING:
				uplet[i].setValue(fields[i].getString());
				break;
			case QTypeTuple.TYPECODE_CHAR:
				uplet[i].setValue(fields[i].getChar());
				break;
			case QTypeTuple.TYPECODE_DATE:
				uplet[i].setValue(fields[i].getDate());
				break;
			case QTypeTuple.TYPECODE_OBJBOOLEAN:
			case QTypeTuple.TYPECODE_OBJBYTE:
			case QTypeTuple.TYPECODE_OBJCHAR:
			case QTypeTuple.TYPECODE_OBJDOUBLE:
			case QTypeTuple.TYPECODE_OBJFLOAT:
			case QTypeTuple.TYPECODE_OBJINT:
			case QTypeTuple.TYPECODE_OBJLONG:
			case QTypeTuple.TYPECODE_OBJSHORT:
				uplet[i].setValue(fields[i].getObject());
				break;
				// TO DO: Add TupleCollection type and Tuple Types
			}
		}
	}

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		if (uplet != null) {
			MemoryTuple mt = (MemoryTuple) clone;
			mt.uplet = new VariableOperand[uplet.length];
			for(int i=0; i<uplet.length; i++) {
				mt.uplet[i] = (VariableOperand) getClone(uplet[i], obj2clone);
			}
		}
		return clone;
	}

	public boolean isDefined(int i) {
		return uplet[i - 1].isDefined();
	}

	public int getSize() {
		return uplet.length;
	}

	public Operand[] toOperandArray() {
		return uplet;
	}

	public int getInt(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getInt();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setInt(int x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public byte getByte(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getByte();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public byte[] getByteArray(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getByteArray();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setByte(byte x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public double getDouble(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getDouble();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setBigDecimal(BigDecimal x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public BigDecimal getBigDecimal(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getBigDecimal();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setBigInteger(BigInteger x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public BigInteger getBigInteger(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getBigInteger();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setDouble(double x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public float getFloat(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				return uplet[i - 1].getFloat();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
	}

	public void setFloat(float x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}


	public short getShort(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getShort();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setShort(short x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public String getString(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getString();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setString(String x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public long getLong(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getLong();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setLong(long x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public Object getObject(int i) throws MedorException {
		Object result = null;
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else {
			try {
				switch (uplet[i - 1].getType().getTypeCode()) {
				case QTypeTuple.TYPECODE_INT:
					result = new Integer(uplet[i - 1].getInt());
					break;
				case QTypeTuple.TYPECODE_SHORT:
					result = new Short(uplet[i - 1].getShort());
					break;
				case QTypeTuple.TYPECODE_BYTE:
					result = new Byte(uplet[i - 1].getByte());
					break;
				case QTypeTuple.TYPECODE_LONG:
					result = new Long(uplet[i - 1].getLong());
					break;
				case QTypeTuple.TYPECODE_DOUBLE:
					result = new Double(uplet[i - 1].getDouble());
					break;
				case QTypeTuple.TYPECODE_BOOLEAN:
					result = new Boolean(uplet[i - 1].getBoolean());
					break;
				case QTypeTuple.TYPECODE_FLOAT:
					result = new Float(uplet[i - 1].getFloat());
					break;
				case QTypeTuple.TYPECODE_STRING:
					String s = uplet[i - 1].getString();
					result = (s == null ? null : new String(s));
					break;
				case QTypeTuple.TYPECODE_CHAR:
					result = new Character(uplet[i - 1].getChar());
					break;
				case QTypeTuple.TYPECODE_DATE:
					result = uplet[i - 1].getDate();
					break;

				default :
					if (uplet[i - 1].getType().getJormName().equalsIgnoreCase("collection")) {
						result = uplet[i - 1].getObject();
						break;
					} else
						result = uplet[i - 1].getObject();
				}
			} catch (TypingException e) {
				throw new MedorException(e);
			}
		}
		return result;
	}

	public boolean getBoolean(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getBoolean();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setBoolean(boolean x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public char getChar(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getChar();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public char[] getCharArray(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getCharArray();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setChar(char x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public Date getDate(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				return uplet[i - 1].getDate();
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public void setDate(Date x, int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			try {
				uplet[i - 1].setValue(x);
			} catch (TypingException e) {
				throw new MedorException(e);
			}
	}

	public TupleCollection getTupleCollection(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			return (TupleCollection) uplet[i - 1].getObject();
	}

	public Operand getLikeOperand(int i) throws MedorException {
		if ((i <= 0) || (i > uplet.length)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			return uplet[i - 1];
	}

}
