/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.optim.lib.DropUnusedProjFieldsRule;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.api.RdbStringQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicRdbStringQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;

public class TestDropUnusedProjFieldRule extends TestMedorHelper {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestDropUnusedProjFieldRule.class);
	}

	public TestDropUnusedProjFieldRule() {
		super("TestDropUnusedProjFieldRule", Log.MEDOR_PREFIX + ".optim.dropunusedprojfield");
	}

	public TestDropUnusedProjFieldRule(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".optim.dropunusedprojfield");
	}


	protected void setUp() {
	}

	protected void tearDown() {
	}

	public void testSimple() {
		QueryTree myRequest = null;
		try {
			RdbStringQueryLeaf ql = new BasicRdbStringQueryLeaf(null, "Select", "");
			RdbField f1 = ql.addRdbField("f1", null, "col1");
			RdbField f2 = ql.addRdbField("f2", null, "col2");
			QueryNode q1 = new Project(new String[]{"f1", "f2"}, ql, "");
			myRequest = new Project(new String[]{"f2"}, q1, "");
		} catch (MedorException e) {
			e.printStackTrace();
			fail("fail to create the request" + e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "Original tree:");
		logger.log(BasicLevel.DEBUG, "==============");
		QueryTreePrinter.printQueryTree(myRequest, logger);

		QueryTree result = null;
		try {
			result = new DropUnusedProjFieldsRule().rewrite(myRequest);
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail("fail to rewrite the request" + e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "Rewritten tree:");
		logger.log(BasicLevel.DEBUG, "===============");
		QueryTreePrinter.printQueryTree(result, logger);

		assertTrue("The root is not a QueryNode", result instanceof QueryNode);
		QueryNode qn = (QueryNode) result;
		QueryTree[] children = qn.getChildren();
		assertNotNull("root.children array is null", children);
		assertEquals("Bad root.children array size", 1, children.length);

		QueryTreeField f = null;
		Field[] previous = null;

		// Check the root node
		try {
			f = (QueryTreeField) qn.getTupleStructure().getField("f1");
			fail("the root.f1 field exists");
		}
		catch (MedorException e) {
		}
		try {
			f = (QueryTreeField) qn.getTupleStructure().getField("f2");
		}
		catch (MedorException e) {
			fail("no field root.f2");
		}
		assertNotNull("no field root.f2", f);
		assertTrue("root.f2 is not a PropagatedField", f instanceof PropagatedField);
		previous = ((PropagatedField) f).getPreviousFields();
		assertNotNull("Previous fields of root.f2 is null", previous);
		assertEquals("Bad number of previous field in root.f2",
			1, previous.length);
		assertTrue("Previous fields of root.f2 is not a QueryTreeField",
			previous[0] instanceof QueryTreeField);
		f = (QueryTreeField) previous[0];
		assertEquals("root.children array or QueryTreeField is wrong",
			children[0], f.getQueryTree());

		// check the middle node
		assertTrue("The middle node is not a QueryNode", children[0] instanceof QueryNode);
		qn = (QueryNode) children[0];
		children = qn.getChildren();
		assertNotNull("middle.children array is null", children);
		assertEquals("Bad middle.children array size", 1, children.length);

		try {
			f = (QueryTreeField) qn.getTupleStructure().getField("f1");
			fail("The middle.f1 field exists");
		}
		catch (MedorException e) {
		}
		try {
			f = (QueryTreeField) qn.getTupleStructure().getField("f2");
		}
		catch (MedorException e) {
			fail("no field root.f2");
		}
		assertNotNull("No field middle.f2", f);
		assertTrue("middle.f2 is not a PropagatedField", f instanceof PropagatedField);
		previous = ((PropagatedField) f).getPreviousFields();
		assertNotNull("Previous fields of middle.f2 is null", previous);
		assertEquals("Bad number of previous field in middle.f2",
			1, previous.length);
		assertTrue("Previous fields of middle.f2 is not a QueryTreeField",
			previous[0] instanceof QueryTreeField);
		f = (QueryTreeField) previous[0];
		assertEquals("middle.children array or QueryTreeField is wrong",
			children[0], f.getQueryTree());

		// check the leaf node
		assertTrue("The middle node is not a RdbStringQueryLeaf",
			children[0] instanceof RdbStringQueryLeaf);
		RdbStringQueryLeaf leaf = (RdbStringQueryLeaf) children[0];
		try {
			f = (QueryTreeField) leaf.getTupleStructure().getField("f1");
		}
		catch (MedorException e) {
			fail("No field leaf.f1");
		}
		assertNotNull("No field leaf.f1", f);
		try {
			f = (QueryTreeField) leaf.getTupleStructure().getField("f2");
		}
		catch (MedorException e) {
			fail("No field leaf.f2");
		}
		assertNotNull("No field leaf.f2", f);
	}

}
