/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.optim.lib.DropUselessNodeRule;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.api.RdbStringQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicRdbStringQueryLeaf;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * It tests the rule DropUselessNodeRule.
 *
 * @author S.Chassande-Barrioz
 */
public class TestDropUselessNodeRule extends TestMedorHelper {

    /**
     * main method to launch the tests manually
     */
    public static void main(String[] args) {
        TestRunner.run(suite());
    }

    /**
     * This method creates a TestSuite object with the current tests
     */
    public static Test suite() {
        return new TestSuite(TestDropUselessNodeRule.class);
    }

    public TestDropUselessNodeRule() {
        super("TestDropUselessNodeRule", Log.MEDOR_PREFIX + ".optim.dropuselessnode");
    }

    public TestDropUselessNodeRule(String testName) {
        super(testName, Log.MEDOR_PREFIX + ".optim.dropuselessnode");
    }


    protected void setUp() {
    }

    protected void tearDown() {
    }

    public void testA() {
        QueryTree myQuery = null;
        try {
            RdbStringQueryLeaf ql = new BasicRdbStringQueryLeaf(null, "Select", "");
            RdbField f1 = ql.addRdbField("f1", PTypeSpace.BYTE, "col1");
            RdbField f2 = ql.addRdbField("f2", PTypeSpace.DATE, "col2");
            QueryNode q1 = new Project(new String[]{"f2"}, ql, "");
            myQuery = new Project(new String[]{"f2"}, q1, "");
        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to create the request" + e.getMessage());
        }
            logger.log(BasicLevel.DEBUG, "Original tree:");
            QueryTreePrinter.printQueryTree(myQuery, logger);

        QueryTree result = null;
        try {
            result = new DropUselessNodeRule().rewrite(myQuery);
        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to rewrite the request" + e.getMessage());
        }
            logger.log(BasicLevel.DEBUG, "Rewritten tree:");
            QueryTreePrinter.printQueryTree(result, logger);
        try {
            RdbStringQueryLeaf ql = new BasicRdbStringQueryLeaf(null, "Select", "");
            RdbField f1 = ql.addRdbField("f1", PTypeSpace.BYTE, "col1");
            RdbField f2 = ql.addRdbField("f2", PTypeSpace.DATE, "col2");
            TestQueryTreeHelper.equals("", ql, result, logger);
        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to create the expected request" + e.getMessage());
        }
    }

    public void testB() {
        QueryTree myQuery = null;
        try {
            RdbStringQueryLeaf ql1 =
                new BasicRdbStringQueryLeaf(null, "Select", "ql1");
            QueryTreeField f1 = ql1.addRdbField("f1", PTypeSpace.LONG, "col1");
            QueryTreeField f2 = ql1.addRdbField("f2", PTypeSpace.CHAR, "col2");
            QueryNode q1 = new Project("q1");
            f1 = q1.addPropagatedField(
                f1.getName(), f1.getType(), new QueryTreeField[]{f1});
            f2 = q1.addPropagatedField(
                f2.getName(), f2.getType(), new QueryTreeField[]{f2});

            RdbStringQueryLeaf ql2 =
                new BasicRdbStringQueryLeaf(null, "Select", "ql2");
            QueryTreeField a = ql2.addRdbField("a", PTypeSpace.LONG, "colA");
            QueryTreeField b = ql2.addRdbField("b", PTypeSpace.CHAR, "colB");
            QueryNode q2 = new Project("q2");
            a = q2.addPropagatedField(
                a.getName(), a.getType(), new QueryTreeField[]{a});
            b = q2.addPropagatedField(
                b.getName(), b.getType(), new QueryTreeField[]{b});

            Expression e = new org.objectweb.medor.expression.lib.And(
                new Equal(new BasicFieldOperand(f1), new BasicFieldOperand(a)),
                new Equal(new BasicFieldOperand(f2), new BasicFieldOperand(b)));
            QueryNode q3 = new JoinProject(e, "q3");
            q3.addPropagatedField(
                a.getName(), a.getType(), new QueryTreeField[]{a});
            q3.addPropagatedField(
                b.getName(), b.getType(), new QueryTreeField[]{b});
            myQuery = q3;

        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to create the request" + e.getMessage());
        }
            logger.log(BasicLevel.DEBUG, "Original tree:");
            QueryTreePrinter.printQueryTree(myQuery, logger);

        QueryTree result = null;
        try {
            result = new DropUselessNodeRule().rewrite(myQuery);
        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to rewrite the request" + e.getMessage());
        }
            logger.log(BasicLevel.DEBUG, "Rewritten tree:");
            QueryTreePrinter.printQueryTree(result, logger);
        try {
            RdbStringQueryLeaf ql1 =
                new BasicRdbStringQueryLeaf(null, "Select", "ql1");
            QueryTreeField f1 = ql1.addRdbField("f1", PTypeSpace.LONG, "col1");
            QueryTreeField f2 = ql1.addRdbField("f2", PTypeSpace.CHAR, "col2");

            RdbStringQueryLeaf ql2 =
                new BasicRdbStringQueryLeaf(null, "Select", "ql2");
            QueryTreeField a = ql2.addRdbField("a", PTypeSpace.LONG, "colA");
            QueryTreeField b = ql2.addRdbField("b", PTypeSpace.CHAR, "colB");
            Expression e = new And(
                new Equal(new BasicFieldOperand(f1), new BasicFieldOperand(a)),
                new Equal(new BasicFieldOperand(f2), new BasicFieldOperand(b)));
            QueryNode q3 = new JoinProject(e, "q3");
            q3.addPropagatedField(
                "q2." + a.getName(), a.getType(), new QueryTreeField[]{a});
            q3.addPropagatedField(
                "q2." + b.getName(), b.getType(), new QueryTreeField[]{b});
                logger.log(BasicLevel.DEBUG, "Expected tree:");
                QueryTreePrinter.printQueryTree(q3, logger);
            TestQueryTreeHelper.equals("", q3, result, logger);
        }
        catch (MedorException e) {
            e.printStackTrace();
            fail("fail to create expected the request" + e.getMessage());
        }

    }

}
