/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim;

import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.datasource.lib.ConnectionFactoryDataStore;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.expression.lib.Plus;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.BasicLevel;
import junit.textui.TestRunner;
import junit.framework.Test;
import junit.framework.TestSuite;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class TestFlattenQueryTreeRule extends TestMedorHelper {

	public final static String LOGGER_NAME =
		Log.MEDOR_PREFIX + ".optim.flattenQueryTree";

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestFlattenQueryTreeRule.class);
	}

	public TestFlattenQueryTreeRule() {
		super("TestFlattenQueryTreeRule", LOGGER_NAME);
	}

	public TestFlattenQueryTreeRule(String testName) {
		super(testName, LOGGER_NAME);
	}

	protected void setUp() {
	}

	protected void tearDown() {
	}

	public void testA() throws MedorException {
		changeLogger(LOGGER_NAME + ".A");
		RdbExpQueryLeaf[] leaves = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject j1 = join("",
			leaves[0], new String[]{"a0"}, leaves[1], new String[]{"a1"});
		JoinProject j2 = join("",
			leaves[2], new String[]{"b2"}, leaves[3], new String[]{"b3"});
		JoinProject init = join("",
			j1, new String[]{"c1"}, j2, new String[]{"c2"});
		init.addCalculatedField("calc", PTypeSpace.LONG, getPlus(j1, "c0", j2, "c3"));
		logger.log(BasicLevel.DEBUG, "Initial query tree:");
		QueryTreePrinter.printQueryTree(init, logger);

		RewriteRule rule = new FlattenQueryTreeRule();
		QueryTree qt = rule.rewrite(init);
		logger.log(BasicLevel.DEBUG, "Rewritten query tree:");
		QueryTreePrinter.printQueryTree(qt, logger);

		RdbExpQueryLeaf[] leaves2 = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject expected = new JoinProject("");
		propagateField(leaves2[0], expected);
		propagateField(leaves2[1], expected);
		propagateField(leaves2[2], expected);
		propagateField(leaves2[3], expected);
		expected.addCalculatedField("calc", PTypeSpace.LONG,
			getPlus(leaves2[0], "c0", leaves2[3], "c3"));
		Expression filter = getEqual(leaves2[1], "c1", leaves2[2], "c2");
		filter = new And(filter, getEqual(leaves2[0], "a0", leaves2[1], "a1"));
		filter = new And(filter, getEqual(leaves2[2], "b2", leaves2[3], "b3"));
		expected.setQueryFilter(filter);
		logger.log(BasicLevel.DEBUG, "Expected query tree:");
		QueryTreePrinter.printQueryTree(expected, logger);

		TestQueryTreeHelper.equals("", expected, qt, logger);
	}

	public void testB() throws MedorException {
		changeLogger(LOGGER_NAME + ".B");
		RdbExpQueryLeaf[] leaves = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject j1 = join("",
			leaves[0], new String[]{"a0"}, leaves[1], new String[]{"a1"});
		JoinProject j2 = join("",
			j1, new String[]{"b1"}, leaves[2], new String[]{"b2"});
		JoinProject init = join("",
			j2, new String[]{"c2"}, leaves[3], new String[]{"c3"});
		init.addCalculatedField("calc", PTypeSpace.LONG, getPlus(j2, "c0", leaves[3], "c3"));

		RewriteRule rule = new FlattenQueryTreeRule();
		QueryTree qt = rule.rewrite(init);
		logger.log(BasicLevel.DEBUG, "Rewritten query tree:");
		QueryTreePrinter.printQueryTree(qt, logger);

		RdbExpQueryLeaf[] leaves2 = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject expected = new JoinProject("");
		propagateField(leaves2[0], expected);
		propagateField(leaves2[1], expected);
		propagateField(leaves2[2], expected);
		propagateField(leaves2[3], expected);
		expected.addCalculatedField("calc", PTypeSpace.LONG,
			getPlus(leaves2[0], "c0", leaves2[3], "c3"));
		Expression filter = getEqual(leaves2[2], "c2", leaves2[3], "c3");
		filter = new And(filter, getEqual(leaves2[1], "b1", leaves2[2], "b2"));
		filter = new And(filter, getEqual(leaves2[0], "a0", leaves2[1], "a1"));
		expected.setQueryFilter(filter);
		logger.log(BasicLevel.DEBUG, "Expected query tree:");
		QueryTreePrinter.printQueryTree(expected, logger);
		TestQueryTreeHelper.equals("", expected, qt, logger);
	}

	public void testC() throws MedorException {
		changeLogger(LOGGER_NAME + ".C");
		RdbExpQueryLeaf[] leaves = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject init = new JoinProject("");
		propagateField(leaves[0], init);
		propagateField(leaves[1], init);
		propagateField(leaves[2], init);
		propagateField(leaves[3], init);
		init.addCalculatedField("calc", PTypeSpace.LONG,
			getPlus(leaves[0], "c0", leaves[3], "c3"));
		Expression filter = getEqual(leaves[2], "c2", leaves[3], "c3");
		filter = new And(filter, getEqual(leaves[1], "b1", leaves[2], "b2"));
		filter = new And(filter, getEqual(leaves[0], "a0", leaves[1], "a1"));
		init.setQueryFilter(filter);

		RewriteRule rule = new FlattenQueryTreeRule();
		QueryTree qt = rule.rewrite(init);
		logger.log(BasicLevel.DEBUG, "Rewritten query tree:");
		QueryTreePrinter.printQueryTree(qt, logger);

		RdbExpQueryLeaf[] leaves2 = createLeaves("rdb", "table", 4,
			new String[]{"a", "b", "c"},
			new PType[]{PTypeSpace.BYTE, PTypeSpace.STRING, PTypeSpace.LONG});
		JoinProject expected = new JoinProject("");
		propagateField(leaves2[0], expected);
		propagateField(leaves2[1], expected);
		propagateField(leaves2[2], expected);
		propagateField(leaves2[3], expected);
		expected.addCalculatedField("calc", PTypeSpace.LONG,
			getPlus(leaves2[0], "c0", leaves2[3], "c3"));
		filter = getEqual(leaves2[2], "c2", leaves2[3], "c3");
		filter = new And(filter, getEqual(leaves2[1], "b1", leaves2[2], "b2"));
		filter = new And(filter, getEqual(leaves2[0], "a0", leaves2[1], "a1"));
		expected.setQueryFilter(filter);
		logger.log(BasicLevel.DEBUG, "Expected query tree:");
		QueryTreePrinter.printQueryTree(expected, logger);
		TestQueryTreeHelper.equals("", expected, qt, logger);
	}

	private JoinProject join(String nodeName,
							 QueryTree qt1, String[] fields1,
							 QueryTree qt2, String[] fields2) throws MedorException {
		JoinProject res = new JoinProject(nodeName);
		join(res, qt1, fields1, qt2, fields2);
		return res;
	}

	private void join(JoinProject join,
					  QueryTree qt1, String[] fields1,
					  QueryTree qt2, String[] fields2) throws MedorException {
		if (fields1.length > 0) {
			Expression filter = join.getQueryFilter();
			for (int i = 0; i < fields1.length; i++) {
				Equal eq = getEqual(qt1, fields1[i], qt2, fields2[i]);
				if (filter == null) {
					filter = eq;
				}
				else {
					filter = new And(filter, eq);
				}
			}
			join.setQueryFilter(filter);
		}
		propagateField(qt1, join);
		propagateField(qt2, join);
	}

	private Equal getEqual(QueryTree qt1, String field1,
						   QueryTree qt2, String field2) throws MedorException {
		Field f1 = qt1.getTupleStructure().getField(field1);
		Field f2 = qt2.getTupleStructure().getField(field2);
		return new Equal(new BasicFieldOperand(f1), new BasicFieldOperand(f2));
	}

	private Plus getPlus(QueryTree qt1, String field1,
						 QueryTree qt2, String field2) throws MedorException {
		Field f1 = qt1.getTupleStructure().getField(field1);
		Field f2 = qt2.getTupleStructure().getField(field2);
		return new Plus(new BasicFieldOperand(f1), new BasicFieldOperand(f2));
	}

	private void propagateField(QueryTree source, QueryNode dest)
		throws MedorException {
		Field[] children = source.getTupleStructure().getFields();
		for (int i = 0; i < children.length; i++) {
			if (!dest.getTupleStructure().contains(children[i].getName())) {
				dest.addPropagatedField(children[i].getName(),
					children[i].getType(),
					new QueryTreeField[]{(QueryTreeField) children[i]});
			}
		}
	}

	private RdbExpQueryLeaf[] createLeaves(String nodeName,
										   String tableName,
										   int nbLeaves,
										   String[] fieldNames,
										   PType[] fieldTypes) throws MedorException {
		DataStore ds = new ConnectionFactoryDataStore(
			DataStore.JDBC_STORE, "postgres", null, this);
		RdbExpQueryLeaf[] res = new RdbExpQueryLeaf[nbLeaves];
		for (int i = 0; i < nbLeaves; i++) {
			String[] fns = new String[fieldNames.length];
			for (int j = 0; j < fieldNames.length; j++) {
				fns[j] = fieldNames[j] + i;
			}
			res[i] = createLeaf(nodeName + i, tableName + i, ds, fns, fieldTypes);
		}
		return res;
	}

	private RdbExpQueryLeaf createLeaf(String nodeName,
									   String tableName,
									   DataStore ds,
									   String[] fieldNames,
									   PType[] fieldTypes) throws MedorException {
		QualifiedTable t = new BasicQualifiedTable(tableName, null);
		RdbExpQueryLeaf res = new BasicRdbExpQueryLeaf(
			ds, new QualifiedTable[]{t}, nodeName);
		for (int i = 0; i < fieldNames.length; i++) {
			res.addRdbField(fieldNames[i], fieldTypes[i], fieldNames[i], t);
		}
		return res;
	}
}
