/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.TestExpressionHelper;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.expression.lib.Greater;
import org.objectweb.medor.expression.lib.LowerEqual;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.optim.lib.PushSelectionRule;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This JUnit test permits to check the PushSelectionRule class.
 *
 * @author S.Chassande-Barrioz
 */
public class TestPushSelectionRule extends TestMedorHelper {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestPushSelectionRule.class);
	}

	public TestPushSelectionRule() {
		super("TestPushSelectionRule", Log.MEDOR_PREFIX + ".optim.pushselection");
	}

	public TestPushSelectionRule(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".optim.pushselection");
	}


	protected void setUp() {
	}

	protected void tearDown() {
	}

	public void testComplex() {
		QueryTreeField fa1 = null,
			fa2 = null,
			fb1 = null,
			fb2 = null,
			fc1 = null,
			fc2 = null,
			fd1 = null,
			fe1 = null,
			fe2 = null;
		QueryNode qn1 = null, root = null;
		BasicRdbExpQueryLeaf ql1 = null, ql2 = null, ql3 = null;
		try {
			QualifiedTable t1 = new BasicQualifiedTable("Table_1", "t1");
			QualifiedTable t2 = new BasicQualifiedTable("Table_2", "t2");
			QualifiedTable t3 = new BasicQualifiedTable("Table_3", "t3");
			ql1 = new BasicRdbExpQueryLeaf(null, new QualifiedTable[] {t1}, "");
			logger.log(BasicLevel.DEBUG, "ql1="+ql1);
			fa1 = ql1.addRdbField("a1", PTypeSpace.LONG, "cola", t1);

			ql2 = new BasicRdbExpQueryLeaf(null, new QualifiedTable[] {t2}, "");
			logger.log(BasicLevel.DEBUG, "ql2="+ql2);
			fb1 = ql2.addRdbField("b1", PTypeSpace.LONG, "colb", t2);
			fc1 = ql2.addRdbField("c1", PTypeSpace.LONG, "colc", t2);
			fe1 = ql2.addRdbField("e1", PTypeSpace.LONG, "cole", t2);

			ql3 = new BasicRdbExpQueryLeaf(null, new QualifiedTable[] {t3}, "");
			logger.log(BasicLevel.DEBUG, "ql3="+ql3);
			fd1 = ql3.addRdbField("d1", PTypeSpace.LONG, "cold", t3);

			qn1 = new JoinProject("");
			logger.log(BasicLevel.DEBUG, "qn1="+qn1);
			fa2 = qn1.addPropagatedField(
				"a2", PTypeSpace.LONG, new QueryTreeField[]{fa1});
			fb2 = qn1.addPropagatedField(
				"b2", PTypeSpace.LONG, new QueryTreeField[]{fb1});
			fc2 = qn1.addPropagatedField(
				"c2", PTypeSpace.LONG, new QueryTreeField[]{fc1});
			fe2 = qn1.addPropagatedField(
				"e2", PTypeSpace.LONG, new QueryTreeField[]{fe1});

			root = new JoinProject("");
			root.addPropagatedField("a3", PTypeSpace.LONG, new QueryTreeField[]{fa2});
			root.addPropagatedField("b3", PTypeSpace.LONG, new QueryTreeField[]{fb2});
			root.addPropagatedField("c3", PTypeSpace.LONG, new QueryTreeField[]{fc2});
			root.addPropagatedField("d2", PTypeSpace.LONG, new QueryTreeField[]{fd1});
			root.addPropagatedField("e3", PTypeSpace.LONG, new QueryTreeField[]{fe2});
		}
		catch (MedorException ex) {
			ex.printStackTrace();
			fail("Impossible to build the tree: " + ex.getMessage());
		}
		PushSelectionRule psr = new PushSelectionRule();
		PushSelectionRule.SameQT sqt = null;
		try {
			//((d2<=3) | (b2>2 & b2=c2)) & ((e2=3 | a2>b2) & a2=12)
			Expression e = null, e2 = null, e3 = null;
			e = new BasicFieldOperand(fd1);
			sqt = psr.isSameQT(e);
			assertTrue("verify isSame d1", sqt.isSame);
			assertEquals("verify qt d1", ql3, sqt.qt);

			e2 = new BasicOperand((long) 3);
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame 3", sqt.isSame);
			assertNull("verify qt 3", sqt.qt);

			e = new LowerEqual(e, e2);
			sqt = psr.isSameQT(e);
			assertTrue("verify isSame d1<=3", sqt.isSame);
			assertEquals("verify qt d1<=3", ql3, sqt.qt);

			e2 = new Greater(new BasicFieldOperand(fb2), new BasicOperand((long) 2));
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame b2>2", sqt.isSame);
			assertEquals("verify qt b2>2", qn1, sqt.qt);
			e3 = new Equal(new BasicFieldOperand(fb2), new BasicFieldOperand(fc2));
			sqt = psr.isSameQT(e3);
			assertTrue("verify isSame b2==c2", sqt.isSame);
			assertEquals("verify qt b2==c2",  qn1, sqt.qt);
			e2 = new And(e2, e3);
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame b2>2 & b2=c2", sqt.isSame);
			assertEquals("verify qt b2>2 & b2=c2",  qn1, sqt.qt);
			e = new Or(e, e2);
			sqt = psr.isSameQT(e);
			assertTrue("verify isSame (d1<=3) | (b2>2 & b2=c2)", !sqt.isSame);

			e2 = new Equal(new BasicFieldOperand(fe2), new BasicOperand((long) 3));
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame e2=3", sqt.isSame);
			assertEquals("verify qt e2=3", qn1, sqt.qt);

			e3 = new Greater(new BasicFieldOperand(fa2), new BasicFieldOperand(fb2));
			sqt = psr.isSameQT(e3);
			assertTrue("verify isSame a2>b2", sqt.isSame);
			assertEquals("verify qt a2>b2", qn1, sqt.qt);

			e2 = new Or(e2, e3);
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame e2=3 | a2>b2", sqt.isSame);
			assertEquals("verify qt e2=3 | a2>b2", qn1, sqt.qt);

			e3 = new Equal(new BasicFieldOperand(fa2), new BasicOperand((long) 12));
			sqt = psr.isSameQT(e3);
			assertTrue("verify isSame a2=12", sqt.isSame);
			assertEquals("verify qt a2=12", qn1, sqt.qt);

			e2 = new And(e2, e3);
			sqt = psr.isSameQT(e2);
			assertTrue("verify isSame (e2=3 | a2>b2) & a2=12", sqt.isSame);
			assertEquals("verify qt (e2=3 | a2>b2) & a2=12", qn1, sqt.qt);

			e = new And(e, e2);
			root.setQueryFilter(e);
			sqt = psr.isSameQT(e);
			assertTrue(
				"verify isSame ((d1<=3) | (b2>2 & b2=c2)) & ((e2=3 | a2>b2) & a2=12)",
				!sqt.isSame);

			// b1>3
			e = new Greater(new BasicFieldOperand(fb1), new BasicOperand((long) 3));
			qn1.setQueryFilter(e);
			sqt = psr.isSameQT(e);
			assertTrue("verify isSame b>3", sqt.isSame);
			assertEquals("verify qt b1>3", ql2, sqt.qt);

			logger.log(BasicLevel.DEBUG, "root.filter=" + e2Str(root.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "qn1.filter=" + e2Str(qn1.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql1.filter=" + e2Str(ql1.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql2.filter=" + e2Str(ql2.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql3.filter=" + e2Str(ql3.getQueryFilter()));
			psr.rewrite(root);
			logger.log(BasicLevel.DEBUG, "root.filter=" + e2Str(root.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "qn1.filter=" + e2Str(qn1.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql1.filter=" + e2Str(ql1.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql2.filter=" + e2Str(ql2.getQueryFilter()));
			logger.log(BasicLevel.DEBUG, "ql3.filter=" + e2Str(ql3.getQueryFilter()));

			//Check root filter: d1<=3 | b2>2 & b2==c2
			//========================================
			Expression compare= null;
			e2 = new LowerEqual(
				new BasicFieldOperand(fd1), new BasicOperand((long) 3));
			e3 = new And(
				new Greater(new BasicFieldOperand(fb2), new BasicOperand((long) 2)),
				new Equal(new BasicFieldOperand(fb2), new BasicFieldOperand(fc2))
				);
			compare = new Or(e2, e3);

			try {
				TestExpressionHelper.equals("root.filter checking :",
					root.getQueryFilter(), compare,	logger);
			}
			catch (MedorException ex) {
				fail("root.filter checking fail: " + ex.getMessage());
			}

			// Check middle.filter: e1=3 | a1>b1
			//==================================
			compare = new Or(
				new Equal(new BasicFieldOperand(fe1), new BasicOperand((long) 3)),
				new Greater(new BasicFieldOperand(fa1), new BasicFieldOperand(fb1)));

			try {
				TestExpressionHelper.equals("qn1.filter checking :",
					qn1.getQueryFilter(), compare, logger);
			}
			catch (MedorException ex) {
				fail("qn1.filter checking fail: " + ex.getMessage());
			}

			// Check ql1.filter: a1 = 12
			//==========================
			compare = new Equal(
				new BasicFieldOperand(fa1),
				new BasicOperand((long) 12));
			try {
				TestExpressionHelper.equals("ql1.filter checking :",
					ql1.getQueryFilter(), compare, logger);
			}
			catch (MedorException ex) {
				fail("ql1.filter checking fail: " + ex.getMessage());
			}

			// Check ql2.filter b > 3
			//=======================
			compare = new Greater(
				new BasicFieldOperand(fb1),
				new BasicOperand((long) 3));
			try {
				TestExpressionHelper.equals("ql2.filter checking :",
					ql2.getQueryFilter(), compare, logger);
			}
			catch (MedorException ex) {
				fail("ql2.filter checking fail: " + ex.getMessage());
			}

			// Check ql3.filter: nothing
			//=======================
			try {
				TestExpressionHelper.equals(
					"ql3.filter checking :", ql3.getQueryFilter(), null,
					logger);
			}
			catch (MedorException ex) {
				fail("ql3.filter  checking fail: " + ex.getMessage());
			}
		}
		catch (MedorException ex) {
			ex.printStackTrace();
			fail("Impossible to do isSameQT: " + ex.getMessage());
		}
	}

	private static final String e2Str(Expression e) {
		return ExpressionPrinter.e2str(e);
	}
}
