/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.query;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.datasource.lib.BasicDataStore;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Greater;
import org.objectweb.medor.expression.lib.Lower;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.filter.lib.Avg;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.optim.rdb.GroupSameDBRule;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.Nest;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * It tests the rule DropUselessNodeRule.
 *
 * @author S.Chassande-Barrioz
 */
public class TestNest extends TestMedorHelper {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestNest.class);
	}

	public TestNest() {
		super("TestNest", Log.MEDOR_PREFIX + ".query.nest");
	}

	public TestNest(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".query.nest");
	}

	protected void setUp() {
	}

	protected void tearDown() {
	}

	public void testA() {
		try {
			logger.log(BasicLevel.DEBUG, "Entering A");
			//create qualified tables
			QualifiedTable t1 = new BasicQualifiedTable("t1", null);
			QualifiedTable t2 = new BasicQualifiedTable("t2", "al2");
			QualifiedTable[] tables1 = {t1, t2};
			// Create a DataStore
			DataStore ds = new BasicDataStore(DataStore.JDBC_STORE, "RDB");
			//create a new leaf
			RdbExpQueryLeaf ql1 =
				new BasicRdbExpQueryLeaf(ds, tables1, "" + "leaf1");
			//add fields
			RdbExpField f1 =
				ql1.addRdbField("f1", PTypeSpace.INT, "col1", t1);
			RdbExpField f2 =
				ql1.addRdbField("f2", PTypeSpace.STRING, "col2", t2);
			//create a filter to be added to the query leaf
			FieldOperand fieldOp1 = new BasicFieldOperand(f1);
			FieldOperand fieldOp2 = new BasicFieldOperand(f2);
			BasicOperand valueL = new BasicOperand(1013);
			BasicOperand valueR = new BasicOperand("coucou");
			Expression left = new Lower(fieldOp1, valueL);
			Expression right = new Greater(fieldOp2, valueR);
			Expression filter1 = new And(left, right);
			//assigne the filter to the query leaf
			ql1.setQueryFilter(filter1);
			// Create a second RdbExpQueryLeaf
			QualifiedTable[] tables2 = {t2, t1};
			RdbExpQueryLeaf ql2 =
				new BasicRdbExpQueryLeaf(ds, tables2, "leaf2");
			RdbExpField f3 =
				ql2.addRdbField("f3", PTypeSpace.STRING, "col3", t2);
			RdbExpField f4 =
				ql2.addRdbField("f4", PTypeSpace.SHORT, "col4", t1);
			//create a filter to be added to the query leaf
			FieldOperand fieldOp3 = new BasicFieldOperand(f3);
			BasicOperand valueR1 = new BasicOperand("ayemayema");
			Expression filter2 = new Equal(fieldOp3, valueR1);
			ql2.setQueryFilter(filter2);
			// Create a JOIN PROJECT Operation
			JoinProject jp = new JoinProject("join");
			QueryTreeField pf5 = jp.addPropagatedField("f5",
			        PTypeSpace.INT,
					new QueryTreeField[]{f1});
			QueryTreeField pf6 = jp.addPropagatedField("f6",
			        PTypeSpace.STRING,
					new QueryTreeField[]{f2, f3});
			QueryTreeField pf7 = jp.addPropagatedField("f7",
			        PTypeSpace.SHORT,
					new QueryTreeField[]{f4});
			// Create a QueryFilter
			FieldOperand fieldOp4 = new BasicFieldOperand(f2);
			FieldOperand fieldOp5 = new BasicFieldOperand(f3);
			Expression filter3 = new Equal(fieldOp4, fieldOp5);
			jp.setQueryFilter(filter3);
			//Create a Nest
			Nest nest = new Nest(
					new QueryTreeField[] {pf5, pf6, pf7},
					"grouped",
					new QueryTreeField[] {},
					"nestNode");
			nest.addCalculatedField("avg",
			        PTypeSpace.FLOAT,
					new Avg(
							new BasicFieldOperand(pf7)
					));
			logger.log(BasicLevel.DEBUG, "Original tree:");
			QueryTreePrinter.printQueryTree(nest, logger);
			logger.log(BasicLevel.DEBUG, ql1.getSqlRequest(null, false, false));
			logger.log(BasicLevel.DEBUG, ql2.getSqlRequest(null, false, false));

			RewriteRule gqr = new GroupSameDBRule();
			RdbExpQueryLeaf rewrittenQuery =
				(RdbExpQueryLeaf) (gqr.rewrite(jp));
			logger.log(BasicLevel.DEBUG, "Rewritten tree:");
			QueryTreePrinter.printQueryTree(rewrittenQuery, logger);

			logger.log(BasicLevel.DEBUG, rewrittenQuery.getSqlRequest(null,
			        false, false));
		}
		catch (Exception e) {
			e.printStackTrace();
			fail("fail to create the request" + e.getMessage());
		}
	}
}
