/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.query;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.TestExpressionHelper;
import org.objectweb.medor.lib.TestMedorHelper;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagFromNestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

public class TestQueryTreeHelper extends TestMedorHelper {

    private static HashSet qtcs = new HashSet();
    
    public static synchronized void registerComparator(QueryTreeComparator qtc) {
        qtcs.add(qtc);
    }
    
	public TestQueryTreeHelper(String testName) {
		super(testName);
	}

    public static void equals(String prefix, QueryTree qt1, QueryTree qt2,
							  Logger log)
		throws MedorException {
		if (qt1 == null && qt2 == null)
			return;
		if (qt1 == null ^ qt2 == null)
			fail(prefix + "one of the query tree instances is null: qt1=" + qt1
				+ " / qt2=" + qt2);
		if (qt1.getDistinct() != qt2.getDistinct())
			fail(prefix + "Not same distinct qt1=" + qt1.getDistinct()
				+ " /qt2=" + qt2.getDistinct());
		if (qt1.getClass() != qt2.getClass())
			fail(prefix + "Not same class qt1=" + qt1.getClass().getName()
				+ " / qt2=" + qt2.getClass().getName());

		log.log(BasicLevel.DEBUG, "QueryNode: name=" + qt1.getName() + " / class: " + qt1.getClass().getName());

		equals(prefix, qt1.getTupleStructure(), qt2.getTupleStructure(), log);

		if (qt1 instanceof QueryNode) {
			QueryNode qn1 = (QueryNode) qt1;
			QueryNode qn2 = (QueryNode) qt2;
			TestExpressionHelper.equals(prefix,
				qn1.getQueryFilter(), qn2.getQueryFilter(), log);
			assertEquals(prefix + "QueryNode(" + qt1.getName() + "): Not same type", qn1.getType(), qn2.getType());
			assertEquals(prefix + "QueryNode(" + qt1.getName() + "): Number of child is different",
				qn1.getChildren().length, qn2.getChildren().length);
			ArrayList child1 = QueryTreePrinter.getChildren(qn1);
			ArrayList child2 = QueryTreePrinter.getChildren(qn2);
			for (int i = 0; i < child1.size(); i++) {
				equals(prefix,
					(QueryTree) child1.get(i),
					(QueryTree) child2.get(i), log);
			}
		} else {
            boolean managed = false;
		    for (Iterator it = qtcs.iterator(); it.hasNext();) {
                QueryTreeComparator qtc = (QueryTreeComparator) it.next();
                managed = qtc.assertEquals(prefix, qt1, qt2, log);
            }
            if (!managed) {
                log.log(BasicLevel.WARN, "Unknown node type: " + qt1);
            }
        }
	}

	public static void equals(String prefix,
							  TupleStructure ts1,
							  TupleStructure ts2,
							  Logger log)
		throws MedorException {
		Field[] f1s = ts1.getFields();
		Field[] f2s = ts2.getFields();
		assertEquals(prefix + "TupleStructure: not same size",
			f1s.length, f2s.length);
		for (int i = 0; i < f1s.length; i++) {
			equals(prefix, f1s[i], f2s[i], log);
		}
	}

	public static void equals(String prefix, Field f1, Field f2, Logger log)
		throws MedorException {
		if (f1 == null && f2 == null)
			return;
		if (f1 == null ^ f2 == null)
			fail(prefix + "one of the fields is null: f1=" + f1
				+ " / f2=" + f2);
		if (f1.getClass() != f2.getClass())
			fail(prefix + "Not same class f1=" + f1.getClass().getName()
				+ " / f2=" + f2.getClass().getName());
		log.log(BasicLevel.DEBUG, "Field class: " + f1.getClass().getName());

		assertEquals(prefix + "Field: Name is different", f1.getName(), f2.getName());
		log.log(BasicLevel.DEBUG, "Field name: " + f1.getName());

		if (f1.getType() != null) {
			log.log(BasicLevel.DEBUG, "Field type f1: " 
                    + f1.getType().getJavaName() 
                    + " / " + f1.getType().getJormName());
			log.log(BasicLevel.DEBUG, "Field type f2: " 
                    + f2.getType().getJavaName() 
                    + " / " + f2.getType().getJormName());
			assertTrue(prefix + "Field: (" + f1.getName() 
                    + ") Type is different (was " + f2.getType().getTypeCode() 
                    + ", expected " + f1.getType().getTypeCode() + ")",
			f1.getType().isa(f2.getType()));
		log.log(BasicLevel.DEBUG, "Field type: " + f1.getType());
		}

		if (f1 instanceof CalculatedField) {
			TestExpressionHelper.equals(prefix,
				((CalculatedField) f1).getExpression(),
				((CalculatedField) f2).getExpression(),
				log);
		}
		else if (f1 instanceof PropagatedField) {
			Field[] fs1 = ((PropagatedField) f1).getPreviousFields();
			Field[] fs2 = ((PropagatedField) f2).getPreviousFields();
			assertEquals(prefix +
				"PropagatedField: Number of previous fields is different for field "
			        + f1.getName() + " of node " + ((PropagatedField) f1).getQueryTree().getName(),
				fs1.length, fs2.length);
			if (f1 instanceof PropagFromNestedField) {
				fs1 = ((PropagFromNestedField) f1).getPreviousFlatField();
				fs2 = ((PropagFromNestedField) f2).getPreviousFlatField();
				assertEquals(prefix +
					"PropagatedField: Number of previous fields is different for field "
				        + f1.getName()  + " of node " + ((PropagatedField) f1).getQueryTree().getName(),
					fs1.length, fs2.length);
			}
		}
        //TODO handle BasicNestedField
		else {
            boolean managed = false;
            for (Iterator it = qtcs.iterator(); it.hasNext();) {
                QueryTreeComparator qtc = (QueryTreeComparator) it.next();
                managed = qtc.assertEquals(prefix, f1, f2, log);
            }
            if (!managed) {
                log.log(BasicLevel.WARN,
                        "Unknown field type: " + f1.getClass().getName());
            }
        }
	}
}
